# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

# python io lib equivalents that provide proper context management, i.e.
# constructors that do not acquire resources and hence do not throw
# (the __enter__ methods acquire resources, e.g. open files)

import os
import fcntl
from typing import overload,cast
import typing
from typing import Literal, Optional, Union
import pathlib
import io
import contextlib
import fcntl
from dataclasses import dataclass
from xju.xn import in_context, in_function_context, first_line_of as l1
from xju.misc import ByteCount
from xju.newtype import Int
from socket import socket, SOCK_NONBLOCK, SOCK_STREAM, SOCK_CLOEXEC, SOL_SOCKET, SO_REUSEADDR
from socket import AF_UNIX
from os import pipe2

class FilePositionTag:pass
class FilePositionBase:pass  # until typing.Self
class FilePosition(FilePositionBase):
    __value:int
    def __init__(self,value:int):
        self.__value=value
        pass
    def value(self):
        return self.__value
    def __str__(self):
        return str(self.__value)
    def __repr__(self):
        return repr(self.__value)
    def __format__(self,f:str)->str:
        return self.__value.__format__(f)
    def __add__(self, x:ByteCount):
        return FilePosition(self.value()+x.value())
    def __radd__(self, x:ByteCount):
        return self.__add__(x)

    @overload
    def __sub__(self, x:FilePositionBase)->ByteCount:
        pass
    @overload
    def __sub__(self, x:ByteCount):  # -> FilePosition
        pass
    def __sub__(self, x):
        if isinstance(x,FilePosition):
            return ByteCount(self.value()-x.value())
        if isinstance(x,ByteCount):
            return FilePosition(self.value()-x.value())
        x_type=type(x)
        assert False, f'cannot subtract {x} of type {x_type} from FilePosition'
    def __int__(self):
        return self.__value
    def __lt__(self, x):
        'compare {self} to {x}'
        try:
            if type(x) is FilePosition:
                return self.value()<x.value()
            raise Exception(f'{x!r} is not a {self.__class__.__name__} it is a {x.__class__.__name__}')
        except Exception as e:
            raise in_function_context(FilePosition.__lt__,vars())
    def __eq__(self, x):
        return not (self<x or x<self)
    def __ne__(self, x):
        return not self==x
    def __le__(self, x):
        return self<x or self==x
    def __ge__(self, x):
        return x<=self
    def __gt__(self, x):
        return x<self
    pass

    
class FileModeBase:  # until typing.Self
    __value:int
    def __init__(self,x:int):
        self.__value=x
        pass
    def value(self):
        return self.__value
    pass

class FileMode(FileModeBase):
    def __str__(self):
        return f'0o{self.value():03o}'
    def __repr__(self):
        return str(self)
    def __add__(self,x:FileModeBase):
        return FileMode(self.value()|x.value())
    def __sub__(self,x:FileModeBase):
        return FileMode(self.value()&(~x.value()))
    def __eq__(self,x):
        if type(x) is FileMode:
            return self.value()==x.value()
        raise Exception(f'{x} is of type {x.__class__} not {self.__class__}')
    def __ne__(self,x):
        if type(x) is FileMode:
            return self.value()!=x.value()
        raise Exception(f'{x} is of type {x.__class__} not {self.__class__}')
    pass

class FileReader(contextlib.AbstractContextManager):
    '''{self.path} reader with close-on-exec {self.close_on_exec}'''
    path:pathlib.Path
    close_on_exec:bool
    input:io.RawIOBase

    def __init__(self, path: pathlib.Path, close_on_exec:bool=True):
        self.path = path
        self.close_on_exec = close_on_exec
        pass
    
    def __str__(self):
        return l1(FileReader.__doc__).format(**vars())

    def __enter__(self):
        '''open {self}'''
        try:
            flags=os.O_RDONLY
            if self.close_on_exec:
                if getattr(os, 'O_CLOEXEC'):
                    flags=flags|os.O_CLOEXEC
            self.__fd = os.open(self.path, flags, 0)
            self.input = io.FileIO(self.__fd, closefd=False)
            return self
        except Exception:
            raise in_function_context(FileReader.__enter__,vars()) from None
        pass

    def __exit__(self, t, e, b):
        '''close {self}'''
        try:
            del self.input
            os.close(self.__fd)
            del self.__fd
        except Exception:
            raise in_function_context(FileReader.__exit__,vars()) from None
        pass

    def seek_to(self, position:FilePosition):
        '''position so next read occurs {position} bytes from start of file
           - returns self'''
        try:
            self.input.seek(int(position), io.SEEK_SET)
            return self
        except Exception:
            raise in_function_context(FileReader.seek_to,vars()) from None
        pass
    
    def seek_by(self, offset:ByteCount):
        '''position {self} so next read occurs {offset} bytes from current position
           - returns self'''
        try:
            self.input.seek(int(offset), io.SEEK_CUR)
            return self
        except Exception:
            raise in_function_context(FileReader.seek_by,vars()) from None
        pass
    
    def size(self) -> ByteCount:
        '''return size of {self}'s file'''
        try:
            return ByteCount(os.fstat(self.__fd).st_size)
        except Exception:
            raise in_function_context(FileReader.size,vars()) from None
        pass

    def position(self) -> FilePosition:
        '''get current position of {self}'''
        try:
            return FilePosition(self.input.seek(0, io.SEEK_CUR))
        except Exception:
            raise in_function_context(FileReader.position,vars()) from None
        pass

    def read(self, read_at_most:ByteCount) -> bytes:
        '''read at most {read_at_most} bytes from current position of {self}'''
        try:
            # files are always blocking, so input.read will never return None
            return cast(bytes,self.input.read(read_at_most.value()))
        except Exception:
            raise in_function_context(FileReader.read,vars()) from None
        pass

    def fd(self) -> int:
        return self.__fd
    pass


class FileWriter(contextlib.AbstractContextManager):
    '''file writer for {self.path} with with create mode {self.mode}, must not exist {self.must_not_exist}, close-on-exec {self.close_on_exec}'''
    path:pathlib.Path
    mode:Optional[FileMode]
    must_not_exist:bool
    close_on_exec:bool
    output:io.RawIOBase

    @overload
    def __init__(self, path: pathlib.Path, mode:FileMode, must_not_exist:Literal[True]=True, close_on_exec:bool=True):
        '''non-existent {path} writer creating with mode {mode}, with close-on-exec {close_on_exec}
           - raises FileExistsError if {path} exists'''
        pass
    @overload
    def __init__(self, path: pathlib.Path, mode:Literal[None]=None, must_not_exist:Literal[False]=False, close_on_exec:bool=True):
        '''existing {path} writer, with close-on-exec {close_on_exec}'''
        pass
    @overload
    def __init__(self, path: pathlib.Path, mode:FileMode, must_not_exist:Literal[False], close_on_exec:bool):
        '''{path} writer creating with mode {mode} if {path} does not exist, with close-on-exec {close_on_exec}'''
        pass
    def __init__(self, path: pathlib.Path, mode=None, must_not_exist=False, close_on_exec=True):
        self.path = path
        self.mode = mode
        self.must_not_exist = must_not_exist
        self.close_on_exec = close_on_exec
        pass

    def __str__(self):
        return l1(FileWriter.__doc__).format(**vars())

    def __enter__(self):
        '''open {self}
           - returns self'''
        try:
            if self.mode is None:
                flags=os.O_WRONLY
            elif self.must_not_exist:
                flags=os.O_WRONLY|os.O_CREAT|os.O_EXCL
            else:
                flags=os.O_WRONLY|os.O_CREAT
            if self.close_on_exec:
                if getattr(os, 'O_CLOEXEC'):
                    flags=flags|os.O_CLOEXEC
                    pass
                pass
            self.__fd = os.open(self.path, flags, int((self.mode or FileMode(0)).value()))
            self.output = io.FileIO(self.__fd, mode='w', closefd=False)
            return self
        except Exception:
            raise in_function_context(FileWriter.__enter__,vars()) from None
        pass
    
    def __exit__(self, t, e, b):
        '''close {self}'''
        try:
            del self.output
            os.close(self.__fd)
            del self.__fd
        except Exception:
            raise in_function_context(FileWriter.__exit__,vars()) from None
        pass
    
    def seek_to(self, position:FilePosition):
        '''position {self} so next write occurs {position} bytes from start of file
           - returns self'''
        try:
            self.output.seek(int(position), io.SEEK_SET)
            return self
        except:
            raise in_function_context(FileWriter.seek_to,vars()) from None
        pass
    
    def seek_by(self, offset:ByteCount):
        '''position {self} so next write occurs {offset} bytes from current position
           - returns self'''
        try:
            self.output.seek(int(offset), io.SEEK_SET)
            return self
        except:
            raise in_function_context(FileWriter.seek_by,vars()) from None
        pass
    
    def size(self) -> ByteCount:
        '''return size of {self}'s file'''
        try:
            return ByteCount(os.fstat(self.__fd).st_size)
        except:
            raise in_function_context(FileWriter.size,vars()) from None
        pass

    def position(self) -> FilePosition:
        '''get current position {self}'''
        try:
            return FilePosition(self.output.seek(0, io.SEEK_CUR))
        except Exception:
            raise in_function_context(FileWriter.position,vars()) from None
        pass

    def fd(self) -> int:
        return self.__fd
    pass

@dataclass
class FileLock(contextlib.AbstractContextManager):
    '''non-blocking "flock" lock {self.target}'''
    target:Union[FileReader,FileWriter]
    def __init__(self, target:Union[FileReader,FileWriter]):
        self.target = target
        pass

    def __str__(self):
        return l1(FileLock.__doc__).format(**vars())

    def __enter__(self):
        '''acquire {self}'''
        try:
            fcntl.flock(self.target.fd(), fcntl.LOCK_EX|fcntl.LOCK_NB)
            return self
        except Exception:
            raise in_function_context(FileLock.__enter__,vars()) from None
        pass

    def __exit__(self, t, e, b):
        '''release {self}'''
        try:
            fcntl.flock(self.target.fd(), fcntl.LOCK_UN)
        except Exception:
            raise in_function_context(FileLock.__exit__,vars()) from None

class UnixStreamListener(contextlib.AbstractContextManager):
    '''{self.path} unix stream socket listener with close-on-exec {self.close_on_exec}'''
    path:pathlib.Path
    close_on_exec:bool
    socket: socket

    def __init__(self, path: pathlib.Path, backlog:int, close_on_exec:bool=True):
        self.path = path
        self.backlog = backlog
        self.close_on_exec = close_on_exec
        pass
    
    def __str__(self):
        return l1(UnixStreamListener.__doc__).format(**vars())

    def __enter__(self):
        '''open {self}'''
        try:
            flags=int(SOCK_NONBLOCK)
            if self.close_on_exec:
                flags=flags|SOCK_CLOEXEC
            s = socket(family=AF_UNIX, type=SOCK_STREAM|flags, proto=-1)
            try:
                s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
                s.bind(str(self.path))
                s.listen(self.backlog)
            except:
                s.close()
                raise
            self.socket=s
            return self
        except Exception:
            raise in_function_context(UnixStreamListener.__enter__,vars()) from None
        pass

    def __exit__(self, t, e, b):
        '''close {self}'''
        try:
            self.socket.close()
            self.path.unlink()
            del self.socket
        except Exception:
            raise in_function_context(UnixStreamListener.__exit__,vars()) from None
        pass
    def fd(self) -> int:
        return self.socket.fileno()
    pass

class UnixStreamSocket(contextlib.AbstractContextManager):
    '''{self.path} unix stream socket with close-on-exec {self.close_on_exec}'''
    path:pathlib.Path
    close_on_exec:bool
    socket:socket
    input:io.RawIOBase
    output:io.RawIOBase

    def __init__(self, x: pathlib.Path|UnixStreamListener, close_on_exec:bool=True):
        self.x=x
        match x:
            case pathlib.Path():
                self.path=x
                pass
            case UnixStreamListener():
                self.path=x.path
                pass
        self.close_on_exec = close_on_exec
    
    def __str__(self):
        return l1(UnixStreamSocket.__doc__).format(**vars())

    def __enter__(self):
        '''make or accept unix stream socket connection'''
        try:
            flags=int(SOCK_NONBLOCK)
            if self.close_on_exec:
                flags=flags|SOCK_CLOEXEC
                pass
            s = socket(family=AF_UNIX, type=SOCK_STREAM|flags, proto=--1)
            try:
                match self.x:
                    case pathlib.Path():
                        try:
                            s.connect(str(self.path))
                        except Exception:
                            raise in_context(f'make connection to {self.path}') from None
                        pass
                    case UnixStreamListener():
                        try:
                            s, addr=self.x.socket.accept()
                        except Exception:
                            raise in_context(f'accept connection on {self.path}') from None
                        pass
                pass
            except:
                s.close()
                raise
            self.socket = s
            self.input = io.FileIO(self.fd(), closefd=False)
            self.output = io.FileIO(self.fd(), mode='w', closefd=False)
            return self
        except Exception:
            raise in_function_context(UnixStreamSocket.__enter__,vars()) from None
        pass

    def __exit__(self, t, e, b):
        '''close {self}'''
        try:
            del self.output
            del self.input
            self.socket.close()
            del self.socket
        except Exception:
            raise in_function_context(UnixStreamSocket.__exit__,vars()) from None
        pass

    def fd(self) -> int:
        return self.socket.fileno()
    pass

class Pipe(contextlib.AbstractContextManager):
    '''pipe with input fd {self.input_fd}, close input on exec {self.close_input_on_exec}, output fd {self.output_fd}, close output on exec {self.close_output_on_exec}'''
    path:pathlib.Path
    close_on_exec:bool
    input_fd: int|None
    output_fd: int|None
    input:io.RawIOBase
    output:io.RawIOBase

    def __init__(self,close_input_on_exec:bool=True,close_output_on_exec:bool=True):
        self.close_input_on_exec=close_input_on_exec
        self.close_output_on_exec=close_output_on_exec
        self.input_fd=None
        self.output_fd=None
    
    def __str__(self):
        return l1(Pipe.__doc__).format(**vars())

    def __enter__(self):
        '''make pipe'''
        try:
            input_fd, output_fd = pipe2(SOCK_NONBLOCK|SOCK_CLOEXEC)
            try:
                if not self.close_input_on_exec:
                    fcntl.fcntl(input_fd, fcntl.FD_CLOEXEC, False)
                if not self.close_output_on_exec:
                    fcntl.fcntl(output_fd, fcntl.FD_CLOEXEC, False)
            except:
                os.close(input_fd)
                os.close(output_fd)
                raise
            self.input_fd = input_fd
            self.output_fd = output_fd
            self.input = io.FileIO(self.input_fd, closefd=False)
            self.output = io.FileIO(self.output_fd, mode='w', closefd=False)
            return self
        except Exception:
            raise in_function_context(Pipe.__enter__,vars()) from None
        pass

    def __exit__(self, t, e, b):
        '''close {self}'''
        try:
            self.close_input()
            self.close_output()
        except Exception:
            raise in_function_context(Pipe.__exit__,vars()) from None
        pass

    def close_input(self) -> None:
        '''close {self} input fd {self.input_fd}'''
        if self.input_fd is not None:
            os.close(self.input_fd)
            self.input_fd=None
            del self.input
            pass
        pass

    def close_output(self) -> None:
        '''close {self} output fd {self.output_fd}'''
        if self.output_fd is not None:
            os.close(self.output_fd)
            self.output_fd=None
            del self.output
            pass
        pass
    pass


# REVISIT: implement algorithms on RawIOBase
# def read(self, read_at_most:ByteCount) -> bytes:
#     '''read at most {read_at_most} bytes from {self}'''
#     try:
#         return self.input.read(read_at_most.value()) or b''
#     except Exception:
#         raise in_function_context(UnixStreamSocket.read,vars()) from None
#     pass

# def write(self, x:bytes) -> int:
#     '''write x or part of it to {self}
#        - return how may bytes were written'''
#     try:
#         return self.output.write(x) or 0
#     except Exception:
#         raise in_function_context(UnixStreamSocket.write,vars()) from None
#     pass

#  and BufferedIOBase
# def read_through(x:bytes) -> bytes
#



# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# One common "execute command" usage, navigating some of the subprocess.
#
import subprocess
from typing import Sequence,Tuple
from xju.xn import in_function_context
import os

class CmdFailed(Exception):
    def __init__(self,argv:Sequence[str],status,stderr:str,stdout:bytes):
        self.argv=argv
        self.status=status
        self.stderr=stderr
        self.stdout=stdout
        Exception.__init__(self,f'non-shell command {argv!r} failed with exit status {status} and stderr {stderr}')
        pass
    pass

def do_cmd(argv:Sequence[str])->Tuple[bytes,  # stdout
                                      str]:   # stderr
    '''do non-shell command {argv!r}'''
    '''returns (stdout,stderr) on zero exit status'''
    '''raises CmdFailed on non-zero exit'''
    try:
        try:
            p=subprocess.Popen(args=argv,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
        except OSError as e:
            if e.errno==2:
                path=os.environ['PATH']
                raise Exception('{argv[0]} not found on PATH {path}'.format(**locals())) from None
            raise
        out,err=p.communicate()
        status=p.returncode
        if status!=0:
            raise CmdFailed(argv,status,err.decode('utf-8'),out)
        return out,err.decode('utf-8')
    except:
        raise in_function_context(do_cmd,vars()) from None
    pass



# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# Time based store.
#
# Provides storage as append-only "bucket" files, each covering a specified number of hours.
# Bucket file content is up to client. Bucket files are append-only to support incremental
# mirroring. Lookup is by time range. Total storage size is constrained by total byte count
# and number of buckets.
#
# Classes TStore, Writer and Reader form the primary interface.
#
from typing import Tuple,Dict,Literal,overload,Sequence,Callable,Any,cast
from xju.cmc.io import FileReader,FileWriter,FileMode,FilePosition
import os
from pathlib import Path
from xju.misc import ByteCount
from xju.time import Hours,Timestamp
from xju.cmc import cmclass,CM
from xju.xn import in_context,in_function_context,first_line_of as l1
import time
from time import struct_time
from calendar import timegm
import io
from bisect import bisect_left,bisect_right
from xju.assert_ import Assert
from xju import jsonschema
import json
from xju.newtype import Int,Str
import shutil

class BucketStartTag:pass
class BucketIDTag:pass

class BucketStart(Int[BucketStartTag]):pass
class BucketID(Str[BucketIDTag]):pass

class NoSuchBucket(Exception):
    bucket_start:BucketStart
    def __init__(self, bucket_start:BucketStart):
        super().__init__(self,f'no bucket starting at {bucket_start}')
        self.bucket_start = bucket_start
        pass
    pass

class BucketExists(Exception):
    bucket_start:BucketStart
    bucket_id:BucketID
    def __init__(self, bucket_start:BucketStart, bucket_id:BucketID):
        super().__init__(self,f'bucket at {bucket_start} exists with uid {bucket_id}')
        self.bucket_start = bucket_start
        self.bucket_id = bucket_id
        pass
    pass

@cmclass
class Reader(CM):
    '''TStore {self.storage_path} reader for bucket with start {self.bucket_start} and id {self.bucket_id}'''
    storage_path:Path
    bucket_start:BucketStart
    bucket_id:BucketID

    __file_reader:FileReader
    
    def __init__(self,
                 storage_path:Path,
                 bucket_start:BucketStart,
                 bucket_id:BucketID,
                 get_bucket_size:Callable[[],ByteCount],
                 file_reader:FileReader):
        self.storage_path=storage_path
        self.bucket_start=bucket_start
        self.bucket_id=bucket_id
        self.__get_bucket_size=get_bucket_size
        self.__file_reader=file_reader
        pass

    def __str__(self):
        return l1(Reader.__doc__).format(**vars())

    def seek_to(self, position:FilePosition):
        '''position TStore reader {self} so next read occurs {position} bytes from start of bucket data
           - returns self'''
        try:
            self.__file_reader.seek_to(position)
            return self
        except:
            raise in_function_context(Reader.seek_to,vars()) from None
        pass
    
    def seek_by(self, offset:ByteCount):
        '''position TStore reader {self} so next read occurs {offset} bytes from current position
           - returns self'''
        try:
            self.__file_reader.seek_by(offset)
            return self
        except:
            raise in_function_context(Reader.seek_by,vars()) from None
        pass
    
    def position(self) -> FilePosition:
        '''get current position'''
        try:
            return self.__file_reader.position()
        except Exception:
            raise in_function_context(Reader.position,vars()) from None
        pass

    def size(self) -> ByteCount:
        '''return size of bucket'''
        try:
            return self.__get_bucket_size()
        except:
            raise in_function_context(Reader.size,vars()) from None
        pass

    def read(self, max_bytes:ByteCount) -> bytes:
        '''read up to {max_bytes} of data from TStore reader {self} current position'''
        try:
            result=self.__file_reader.read(max_bytes)
            return result
        except Exception:
            raise in_function_context(Reader.read,vars()) from None
        pass
    pass


@cmclass
class Writer(CM):
    '''writer for TStore {self.storage_path}' bucket with start {self.bucket_start} and id {self.bucket_id}'''
    storage_path:Path
    bucket_start:BucketStart
    bucket_id:BucketID

    __file_writer:FileWriter
    
    def __init__(self,
                 storage_path:Path,
                 bucket_start:BucketStart,
                 bucket_id:BucketID,
                 get_bucket_size:Callable[[],ByteCount],
                 get_room:Callable[[],ByteCount],
                 file_writer:FileWriter,
                 appended:Callable[[ByteCount],Any]):
        self.storage_path=storage_path
        self.bucket_start=bucket_start
        self.bucket_id=bucket_id
        self.__get_bucket_size=get_bucket_size
        self.__get_room=get_room
        self.__file_writer=file_writer
        self.__appended=appended
        pass
    
    def size(self) -> ByteCount:
        '''return size of bucket'''
        try:
            return self.__get_bucket_size()
        except Exception:
            raise in_function_context(Writer.size,vars()) from None
        pass

    def append(self, data:bytes):
        '''TStore writer {self} append {len_data} bytes of data assuming there is room
           - assumes our bucket still in tstore
           - raises NoSuchBucket if bucket no longer exists
           - raises BucketExists if bucket now has different id'''
        len_data=len(data)
        try:
            Assert(self.__get_room()>=ByteCount(len(data)))
            self.__file_writer.seek_to(FilePosition(self.size().value()))
            self.__file_writer.output.write(data)
            self.__appended(ByteCount(len(data)))
        except Exception:
            raise in_function_context(Writer.append,vars()) from None
        pass
    pass

class TStore:
    '''{self.storage_path} ({self.hours_per_bucket} hours per bucket, max {self.max_buckets} buckets, max {self.max_size} bytes and file creation mode {self.file_creation_mode})'''
    storage_path:Path
    hours_per_bucket:Hours
    max_buckets: int
    max_size:ByteCount
    file_creation_mode:FileMode
    
    __buckets:Dict[BucketStart,BucketID]
    __bucket_sizes:Dict[Tuple[BucketStart,BucketID],ByteCount]
    __current_size:ByteCount

    def __str__(self):
        return l1(TStore.__doc__).format(**vars())

    def current_size(self) -> ByteCount:
        return self.__current_size

    def get_bucket_sizes(self) -> Dict[Tuple[BucketStart,BucketID],ByteCount]:
        return self.__bucket_sizes.copy()

    @overload
    def __init__(self,
                 storage_path:Path,
                 hours_per_bucket:Hours,
                 max_buckets: int,
                 max_size:ByteCount,
                 file_creation_mode:FileMode):
        '''create non-existent TStore at {storage_path} with mode {file_creation_mode}, {hours_per_bucket} hours per bucket, {max_buckets} buckets max,  {max_size} total bytes max
           - hours_per_bucket must be a factor of 24
           - raises FileExistsError if TStore exists'''

    @overload
    def __init__(self,
                 storage_path: Path):
        '''open existing TStore at {storage_path}
           - raises FileNotFoundError if TStore does not exist'''

    def __init__(self,
                 storage_path:Path,
                 hours_per_bucket=None,
                 max_buckets=None,
                 max_size=None,
                 file_creation_mode=None):
        self.storage_path=storage_path
        if (isinstance(hours_per_bucket,Hours) and
            isinstance(max_buckets,int) and
            isinstance(max_size,ByteCount) and
            isinstance(file_creation_mode,FileMode)):
            try:
                storage_path.mkdir(mode=file_creation_mode.value())
                try:
                    assert hours_per_bucket.value() in [1,2,3,4,6,8,12,24] # factor of 24
                    self.hours_per_bucket, self.max_buckets, \
                        self.max_size, self.file_creation_mode=write_attrs(storage_path,
                                                                           hours_per_bucket,
                                                                           max_buckets,
                                                                           max_size,
                                                                           file_creation_mode)
                    self.__buckets={}
                    self.__bucket_sizes={}
                    self.__current_size=ByteCount(0)
                except Exception:
                    shutil.rmtree(str(storage_path))
                    raise
            except Exception as e:
                raise in_context('create non-existent TStore at {storage_path} with mode {file_creation_mode}, {hours_per_bucket} hours per bucket, max buckets {max_buckets}, max size {max_size} bytes'.format(**vars())) from None
            pass
        else:
            try:
                self.hours_per_bucket, self.max_buckets, \
                    self.max_size, self.file_creation_mode=read_attrs(storage_path)
                self.__bucket_sizes = read_files(storage_path)
                self.__buckets={bucket_start:bucket_id
                                for bucket_start,bucket_id in self.__bucket_sizes}
                self.__current_size=sum(self.__bucket_sizes.values(),ByteCount(0))
            except Exception as e:
                raise in_context('open existing TStore at {storage_path}'.format(**vars())) from None
            pass
        pass
            
    def calc_bucket_start(self, timestamp:Timestamp) -> BucketStart:
        '''calculate start of TStore {self}'s bucket for timestamp {timestamp}
           - bucket need not exist'''
        x = time.gmtime(float(timestamp))
        h = self.hours_per_bucket*(Hours(x.tm_hour)//self.hours_per_bucket)
        return BucketStart(int(timegm(struct_time( (x.tm_year,x.tm_mon,x.tm_mday,int(h),0,0,0,0,0) ))))
        
    def get_buckets_of(self, begin:Timestamp, end:Timestamp) -> Sequence[Tuple[BucketStart,BucketID]]:
        '''get starts and ids of TStore {self}'s existing buckets covering time range [{begin},{end})'''
        begin_bucket=self.calc_bucket_start(begin)
        end_bucket=self.calc_bucket_start(end)
        all_buckets=[(bucket_start,bucket_id) for bucket_start,bucket_id in self.__buckets.items()]
        all_buckets.sort()
        all_starts=[bucket_start for bucket_start,bucket_id in all_buckets]
        return all_buckets[bisect_left(all_starts,begin_bucket):bisect_right(all_starts,end_bucket)]

    def list_unseen(self,seen:Dict[Tuple[BucketStart,BucketID], ByteCount]) -> Dict[Tuple[BucketStart,BucketID], ByteCount]:
        '''list sizes of TStore {self} buckets with unseen data having seen {seen}
           - note where an item in seen is past the end of our current
             data, result will include that bucket with its current size
           - where seen item does not exist (e.g. any more) it will appear in result 
             with size 0'''
        result:Dict[Tuple[BucketStart,BucketID], ByteCount]={}
        seen_keys=set(seen.keys())
        bucket_keys=set([(bucket_start,bucket_id) for bucket_start,bucket_id in self.__buckets.items()])
        # no longer exist
        for bucket_start,bucket_id in seen_keys-bucket_keys:
            result[(bucket_start,bucket_id)]=ByteCount(0)
            pass
        # still exist
        for bucket_start,bucket_id in set.intersection(seen_keys,bucket_keys):
            current_size=self.__bucket_sizes[(bucket_start,bucket_id)]
            if current_size and (current_size != seen[(bucket_start,bucket_id)]):
                result[(bucket_start,bucket_id)]=current_size
                pass
            pass
        # new
        for bucket_start,bucket_id in bucket_keys-seen_keys:
            current_size=self.__bucket_sizes[(bucket_start,bucket_id)]
            if current_size:
                result[(bucket_start,bucket_id)]=current_size
                pass
            pass
        return result

    def get_bucket(self,bucket_start:BucketStart) -> BucketID:
        '''get {self} existing bucket with start {bucket_start}
           - raises NoSuchBucket if bucket does not exist'''
        if bucket_start in self.__buckets:
            return self.__buckets[bucket_start]
        raise NoSuchBucket(bucket_start)

    def verify_bucket_exists(self,bucket_start:BucketStart,bucket_id:BucketID):
        '''verify that bucket with start {bucket_start} exists and has id {bucket_id}'''
        try:
            if bucket_start not in self.__buckets:
                raise NoSuchBucket(bucket_start)
            if (bucket_start,bucket_id) not in self.__bucket_sizes:
                raise BucketExists(bucket_start,self.__buckets[bucket_start])
            pass
        except Exception:
            raise in_function_context(TStore.verify_bucket_exists,vars()) from None
        pass
    
    def create_bucket(self,bucket_start:BucketStart,bucket_id:BucketID):
        '''create {self} bucket with start {bucket_start} and id {bucket_id}
           - bucket can then be read and writtern with Reader and Writer
           - raises BucketExists if bucket already exists with specified start
           - deletes bucket with oldest start to make room if already at max_buckets'''
        try:
            if bucket_start in self.__buckets:
                raise BucketExists(bucket_start,self.__buckets[bucket_start])
                pass
            self.__trim_buckets(self.max_buckets-1)
            path=get_path_of(self.storage_path,bucket_start,bucket_id,self.hours_per_bucket)

            make_path_dirs(path,self.storage_path,self.file_creation_mode,True)
            with FileWriter(path,
                            mode=self.file_creation_mode-FileMode(0o111),
                            must_not_exist=True):
                pass
            self.__buckets[bucket_start]=bucket_id
            self.__bucket_sizes[(bucket_start,bucket_id)]=ByteCount(0)
        except Exception as e:
            raise in_function_context(TStore.create_bucket,vars())
        pass

    def make_room_for(self, byte_count:ByteCount) -> None:
        '''make room in TStore {self} for {byte_count} bytes of data
           - deletes just enough buckets with oldest start times to make room'''
        Assert(byte_count)<self.max_size
        self.__trim_bytes(self.max_size-byte_count)
        pass

    def delete_bucket(self, bucket_start:BucketStart,bucket_id:BucketID):
        '''delete {self} bucket with start {bucket_start} and id {bucket_id}'''
        try:
            bucket_size=self.__bucket_sizes[(bucket_start,bucket_id)]
            path=get_path_of(self.storage_path,bucket_start,bucket_id,self.hours_per_bucket)
            path.unlink()
            path=path.parent
            while path != self.storage_path and len(list(path.iterdir()))==0:
                path.rmdir()
                path=path.parent
                pass
            self.__current_size=self.__current_size-bucket_size
            del self.__buckets[bucket_start]
            del self.__bucket_sizes[(bucket_start,bucket_id)]
        except Exception:
            raise in_function_context(TStore.delete_bucket,vars()) from None
        pass
            
    def __trim_buckets(self, max_buckets):
        '''trim TStore {self}'s buckets so that there are at most {max_buckets} of them
           - deletes buckets with oldest start'''
        try:
            Assert(max_buckets)>0
            if len(self.__buckets)>max_buckets:
                all_buckets=list(self.__buckets.items())
                all_buckets.sort()
                i=0
                while len(self.__buckets)>max_buckets and i<len(all_buckets):
                    self.delete_bucket(*all_buckets[i])
                    i+=1
                    pass
                pass
            pass
        except Exception as e:
            raise in_function_context(TStore.__trim_buckets,vars()) from None
        pass

    def __trim_bytes(self,ensure_at_most:ByteCount):
        '''trim {self} storage to ensure it is at most {ensure_at_most} bytes'''
        try:
            if self.__current_size > ensure_at_most:
                ordered=list(self.__bucket_sizes)
                ordered.sort()
                i=0
                while self.__current_size > ensure_at_most and i<len(ordered):
                    self.delete_bucket(*ordered[i])
                    i+=1
                pass
            pass
        except Exception as e:
            raise in_function_context(TStore.__trim_bytes,vars()) from None
        pass

    def get_bucket_size(self,bucket_start:BucketStart,bucket_id:BucketID)->ByteCount:
        return self.__bucket_sizes[(bucket_start,bucket_id)]

    def new_reader(self,bucket_start:BucketStart, bucket_id:BucketID)->Reader:
        '''create reader for TStore {self}'s bucket with start {bucket_start} and id {bucket_id}
           - raises NoSuchBucket if such a bucket does not yet exist
           - raises BucketExists if bucket with start {bucket_start} has different id'''
        try:
            self.verify_bucket_exists(bucket_start,bucket_id)
            def get_bucket_size():
                return self.get_bucket_size(bucket_start,bucket_id)
            return Reader(
                self.storage_path,
                bucket_start,
                bucket_id,
                get_bucket_size,
                FileReader(get_path_of(self.storage_path,bucket_start,bucket_id,self.hours_per_bucket)))
        except Exception as e:
            raise in_function_context(TStore.new_reader,vars()) from None
        pass

    def new_writer(self,bucket_start:BucketStart, bucket_id:BucketID)->Writer:
        '''create writer for TStore {self}'s bucket with start {bucket_start} and id {bucket_id}
           - bucket must exist
           - ensures store stays within max_buckets along the way'''
        try:
            self.verify_bucket_exists(bucket_start,bucket_id)
            def get_bucket_size():
                return self.get_bucket_size(bucket_start,bucket_id)
            def get_room():
                return self.max_size-self.current_size()
            def appended(number_of_bytes:ByteCount):
                self.__current_size=self.__current_size+number_of_bytes
                self.__bucket_sizes[(bucket_start,bucket_id)]=\
                    self.__bucket_sizes[(bucket_start,bucket_id)]+number_of_bytes
                pass
            return Writer(
                self.storage_path,
                bucket_start,
                bucket_id,
                get_bucket_size,
                get_room,
                FileWriter(
                    get_path_of(self.storage_path,bucket_start,bucket_id,self.hours_per_bucket)),
                appended)
        except Exception as e:
            raise in_function_context(TStore.new_writer,vars()) from None
        pass
    pass


attrs_schema=jsonschema.Schema({
    'hours_per_bucket':int,
    'max_buckets':int,
    'max_size':int,
    'file_creation_mode':int})

TSTORE_ATTRS='tstore.json'

def read_attrs(storage_path:Path,
               attrs_file=TSTORE_ATTRS) -> Tuple[Hours, int, ByteCount, FileMode]:
    '''read TStore attributes from {storage_path}/{attrs_file}'''
    try:
        with FileReader(storage_path/attrs_file) as f:
            x=json.loads(f.input.read().decode('utf-8'))
            attrs_schema.validate(x)
            hours_per_bucket=x['hours_per_bucket']
            max_buckets=x['max_buckets']
            max_size=x['max_size']
            file_creation_mode=x['file_creation_mode']
            Assert(hours_per_bucket).isInstanceOf(int)
            Assert(max_buckets).isInstanceOf(int)
            Assert(max_size).isInstanceOf(int)
            Assert(file_creation_mode).isInstanceOf(int)
            pass
        return (Hours(cast(int,hours_per_bucket)),
                cast(int,max_buckets),
                ByteCount(cast(int,max_size)),
                FileMode(cast(int,file_creation_mode)))
    except Exception as e:
        raise in_function_context(read_attrs,vars()) from None
    pass

def write_attrs(storage_path:Path,
                hours_per_bucket:Hours,
                max_buckets:int,
                max_size:ByteCount,
                file_creation_mode:FileMode,
                attrs_file=TSTORE_ATTRS) -> Tuple[Hours,int,ByteCount,FileMode]:
    '''write TStore attrs hours per bucket {hours_per_bucket}, ''' \
        '''max_buckets {max_buckets}, max size {max_size} bytes, file creation mode ''' \
        '''{file_creation_mode} to {storage_path}/{attrs_file}'''
    try:
        attrs=attrs_schema.validate({
            'hours_per_bucket':hours_per_bucket.value(),
            'max_buckets':max_buckets,
            'max_size':max_size.value(),
            'file_creation_mode':file_creation_mode.value()})
        with FileWriter(storage_path/attrs_file,
                        must_not_exist=True,
                        mode=file_creation_mode-FileMode(0o111)) as f:
            f.output.write(json.dumps(attrs).encode('utf-8'))
            pass
        return (hours_per_bucket,max_buckets,max_size,file_creation_mode)
    except Exception as e:
        raise in_function_context(write_attrs,vars()) from None
    pass

def get_path_of(storage_path:Path,
                bucket_start:BucketStart,
                bucket_id:BucketID,
                hours_per_bucket:Hours) -> Path:
    '''get path to {storage_path} bucket with start {bucket_start} and id {bucket_id} ''' \
        '''where storage has {hours_per_bucket} hours per bucket'''
    '''- bucket need not exist'''
    x = time.gmtime(int(bucket_start))
    h = hours_per_bucket*(Hours(x.tm_hour)//hours_per_bucket)
    s = struct_time( (x.tm_year,x.tm_mon,x.tm_mday,int(h),0,0,0,0,0) )
    return storage_path / f'{s.tm_year}'/f'{s.tm_mon:02}'/f'{s.tm_mday:02}'/f'{s.tm_hour:02}'/f'{bucket_id}.txt'

def read_files(storage_path:Path) -> Dict[Tuple[BucketStart,BucketID],ByteCount]:
    '''read tstore files in {storage_path}'''
    try:
        result={}
        for ydir in [y for y in storage_path.iterdir() if y.is_dir()]:
            year=int(ydir.name)
            try:
                for mdir in [m for m in ydir.iterdir() if m.is_dir()]:
                    month=int(mdir.name)
                    try:
                        for ddir in [d for d in mdir.iterdir() if d.is_dir()]:
                            mday=int(ddir.name)
                            try:
                                for hdir in [h for h in ddir.iterdir() if h.is_dir()]:
                                    hour=int(hdir.name)
                                    try:
                                        bucket_start=BucketStart(timegm(struct_time(
                                            (year,month,mday,hour,0,0,0,0,0))))
                                        bucket_ids=[BucketID(b.name).removesuffix('.txt')
                                                    for b in hdir.iterdir()
                                                    if b.is_file()]
                                        if len(bucket_ids)>1:
                                            raise Exception(
                                                f'expected at most one bucket data file, not {bucket_ids}')
                                        if len(bucket_ids):
                                            bucket_id=bucket_ids[0]
                                            result[(bucket_start,bucket_id)]=ByteCount(
                                                (hdir/f'{bucket_id}.txt').stat().st_size)
                                            pass
                                        pass
                                    except Exception:
                                        raise in_context('stat bucket files in hour {hdir.name} directory'.format(**vars()))
                                    pass
                                pass
                            except Exception:
                                raise in_context(
                                    f'read hour directories in mday {ddir.name} directory'.format(**vars()))
                            pass
                        pass
                    except Exception:
                        raise in_context(f'read day directories in month {mdir.name} directory'.format(**vars()))
                    pass
                pass
            except Exception:
                raise in_context(f'read month directories in year {ydir.name} directory'.format(**vars()))
            pass
        return result
    except Exception:
        raise in_function_context(read_files,vars())
    pass

def make_path_dirs(path:Path, root:Path, mode:FileMode, exist_ok:bool):
    '''make directories of {path} that are not in {root} (assumed to be a parent directory)'''
    try:
        rel=path.relative_to(root)
        for d in reversed(rel.parents):
            try:
                (root/d).mkdir(mode=mode.value(),exist_ok=exist_ok)
            except Exception as e:
                raise in_context(f'make directory {root/d}') from None
            pass
        pass
    except Exception:
        raise in_function_context(make_path_dirs,vars()) from None
    pass



# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# perflog stores time-series data in a size-limited structured directory of plain text
# files suited to low-rate writes (e.g. 1 record every 5 seconds) and retrieval by
# time period, with each record being a sequence of optional int/float/str/bool/json values
# to a specified schema.
#
# perflog supports incremental mirroring of data (one-way-sync), e.g. supporting
# incremental remote backup/mirroring
#
# Classes PerfLog, Recorder and Tracker form the primary interface.
#

from pathlib import Path
from typing import Union, Literal,List,Optional,Dict,Tuple,overload,Callable,Any,Generator,cast
from xju.misc import ByteCount
from xju.xn import in_context,in_function_context,first_line_of as l1
from xju.cmc.tstore import TStore,Reader,Writer,BucketStart,BucketID,NoSuchBucket,BucketExists
from xju.cmc.io import FilePosition,FileMode
from xju.time import Timestamp,now,Hours,Duration
from xju.cmc import cmclass,CM,Dict as CMDict
from xju.cmc.io import FileReader,FileWriter
from xju.newtype import Str
from xju.assert_ import Assert
from xju import jsonschema
from xju.jsonschema import OneOf
import os
import json
import io
import shutil

class ColNameType:pass
class ColName(Str[ColNameType]):pass

ColType=Literal[
    'str','int','float','bool',          # not allowed to be None
    '(str)','(int)','(float)','(bool)',  # optional, i.e. value can be None
    'json'                               # any json-ifiable value
]

# mypy can't do isinstance on instance of generic (even though all types
# involved are concrete; workaround is to use a concrete subclass
class Writers(CMDict[ Tuple[BucketStart,BucketID], Writer ]):pass

@cmclass
class Recorder(CM):
    '''recorder recording records to PerfLog {self.storage_path}'''
    storage_path:Path
    schema:Dict[ColName,ColType]

    __tstore:TStore
    __writers:Writers

    def __init__(self,
                 storage_path:Path,
                 schema:Dict[ColName,ColType],
                 tstore:TStore):
        self.storage_path=storage_path
        self.schema=schema
        self.__tstore=tstore
        self.__writers=Writers()
        pass

    def __str__(self):
        return l1(Recorder.__doc__).format(**vars())

    def record(self,timestamp:Timestamp, record:List):
        '''{self} record at timestamp {timestamp} record {record!r}'''
        try:
            bucket_start=self.__tstore.calc_bucket_start(timestamp)
            encoded_record=encode_timestamped_record(timestamp-Timestamp(bucket_start.value()),
                                                     record,
                                                     self.schema)
            self.__tstore.make_room_for(ByteCount(len(encoded_record)))
            try:
                bucket_id=self.__tstore.get_bucket(bucket_start)
            except NoSuchBucket:
                t=now()
                bucket_id=BucketID(f'{t:0.06f}')
                self.__tstore.create_bucket(bucket_start,bucket_id)
                pass
            if (bucket_start,bucket_id) not in self.__writers:
                # only keep one writer because most records will go into same or new bucket
                self.__writers.clear()
                self.__writers[(bucket_start,bucket_id)]=self.__tstore.new_writer(
                    bucket_start,
                    bucket_id)
                pass
            self.__writers[(bucket_start,bucket_id)].append(encoded_record)
        except Exception:
            raise in_function_context(Recorder.record,vars())
        pass
    pass


class Tracker:
    '''tracker updating PerfLog {self.storage_path} with unseen data
       - tracker assumes PerfLog has same max bytes and max buckets as source'''
    storage_path:Path
    schema:Dict[ColName,ColType]

    __tstore:TStore

    def __init__(self,
                 storage_path:Path,
                 tstore:TStore):
        self.storage_path=storage_path
        self.__tstore=tstore
        pass

    def __str__(self):
        return l1(Tracker.__doc__).format(**vars())

    def get_seen(self) -> Dict[Tuple[BucketStart,BucketID], ByteCount]:
        return self.__tstore.get_bucket_sizes()

    def write_unseen_data(self, unseen:Dict[Tuple[BucketStart,BucketID],
                                            Tuple[FilePosition,bytes]]):
        '''{self} write unseen data to PerfLog {self.storage_path}
           - caller must ensure that unseen data with non-zero position matches
             current data size for that bucket
           - caller must ensure that underlying tstore max_size, max_buckets and hours_per_bucket
             matches that of the source data'''
        try:
            writer:Writer
            for bucket_start, bucket_id in unseen:
                position,data=unseen[(bucket_start,bucket_id)]
                data_len=len(data)
                try:
                    try:
                        existing_id=self.__tstore.get_bucket(bucket_start)
                        if existing_id != bucket_id:
                            self.__tstore.delete_bucket(bucket_start,existing_id)
                            self.__tstore.get_bucket(bucket_start) # will raise
                    except NoSuchBucket:
                        self.__tstore.create_bucket(bucket_start,bucket_id)
                        pass
                    if position==FilePosition(0):
                        self.__tstore.delete_bucket(bucket_start,bucket_id)
                    if len(data):
                        try:
                            self.__tstore.create_bucket(bucket_start,bucket_id)
                        except BucketExists:
                            pass
                        # there is room (as long as __tstore has same max_bytes, max_buckets
                        # as source)
                        with self.__tstore.new_writer(bucket_start,bucket_id) as writer:
                            Assert(FilePosition(0)+writer.size())==position
                            writer.append(data)
                        pass
                    pass
                except Exception:
                    raise in_context(f'write {data_len} bytes of data at position {position} '
                                     f'of bucket with start {bucket_start} and id {bucket_id}')
                pass
            pass
        except Exception:
            raise in_function_context(Tracker.write_unseen_data,vars()) from None
        pass


class PerfLog:
    '''{self.storage_path} size limited to {self.max_size} bytes/{self.max_buckets} files with ''' \
        '''record schema {self.schema}, {self.hours_per_bucket} hours per file and ''' \
        '''filecreation mode {self.file_creation_mode}'''
    storage_path:Path
    schema:Dict[ColName,ColType]
    hours_per_bucket:Hours
    max_size:ByteCount
    max_buckets:int
    file_creation_mode:FileMode
    
    __tstore:TStore

    @overload
    def __init__(self,
                 storage_path:Path,
                 schema:Dict[ColName,ColType],
                 hours_per_bucket:Hours,
                 max_buckets: int,
                 max_size:ByteCount,
                 file_creation_mode:FileMode):
        '''create non-existent PerfLog with schema {schema} at {storage_path} with mode ''' \
        '''{file_creation_mode}, {hours_per_bucket} hours per bucket, max buckets ''' \
        '''{max_buckets}, max size {max_size} bytes
           - hours_per_bucket must be a factor of 24
           - raises FileExistsError if perflog exists at {storage_path}'''

    @overload
    def __init__(self,storage_path:Path):
        '''open existing PerfLog at {storage_path} reading attributes from its perflog.json file
           - raises FileNotFoundError if PerfLog does not exist'''

    def __init__(self,
                 storage_path:Path,
                 schema:Optional[Dict[ColName,ColType]]=None,
                 hours_per_bucket:Optional[Hours]=None,
                 max_buckets:Optional[int]=None,
                 max_size:Optional[ByteCount]=None,
                 file_creation_mode:Optional[FileMode]=None):
        self.storage_path = storage_path
        if (isinstance(schema,dict) and
            isinstance(hours_per_bucket,Hours) and
            isinstance(max_buckets, int) and
            isinstance(max_size,ByteCount) and
            isinstance(file_creation_mode,FileMode)):
            try:
                self.schema=schema
                self.hours_per_bucket=hours_per_bucket
                self.max_size=max_size
                self.max_buckets=max_buckets
                self.file_creation_mode=file_creation_mode
                storage_path.mkdir(mode=file_creation_mode.value())
                try:
                    write_attrs(storage_path, schema, file_creation_mode)
                    self.__tstore=TStore(storage_path / 'tstore',
                                         hours_per_bucket,
                                         max_buckets,
                                         max_size,
                                         file_creation_mode)
                except Exception:
                    shutil.rmtree(storage_path)
                    raise
                pass
            except Exception as e:
                raise in_context(
                    '''create non-existent PerfLog with schema {schema} at {storage_path} with ''' \
                    '''mode {file_creation_mode}, {hours_per_bucket} hours per bucket, max ''' \
                    '''buckets {max_buckets}, max size {max_size} bytes'''.format(**vars())) from None
            pass
        else:
            try:
                self.__tstore=TStore(storage_path / 'tstore')
                self.hours_per_bucket=self.__tstore.hours_per_bucket
                self.max_size=self.__tstore.max_size
                self.max_buckets=self.__tstore.max_buckets
                self.file_creation_mode=self.__tstore.file_creation_mode
                self.schema=read_attrs(storage_path)
            except Exception as e:
                raise in_context(
                    '''open existing PerfLog at {storage_path} reading attributes from its perflog.json file'''.format(**vars())) from None
            pass
        pass

    def __str__(self) -> str:
        return l1(PerfLog.__doc__).format(**vars())
    
    def fetch(self,
              begin:Timestamp,
              end:Timestamp,
              max_records:int,
              max_bytes:ByteCount,
              corruption_handler:Callable[[Exception],Any]) -> Generator[
                  Tuple[float, List[Union[str,int,float,bool,None,List,Dict]]],None,None]:
        '''yield each record of PerfLog {self} with timestamp at or after {begin} excluding records with timestamp at or after {end} and all further records once {max_bytes} bytes have already been yielded or {max_records} records have been yielded
           - note that records returned are in addition order, which is not necessarily time order
           - note max_bytes is applied to the pre-decode (i.e. stored) record data'''
        try:
            records_yielded = 0
            bytes_yielded = ByteCount(0)
            r:Reader
            for bucket_start,bucket_id in self.__tstore.get_buckets_of(begin, end):
                try:
                    with self.__tstore.new_reader(bucket_start,bucket_id) as r:
                        ts=Timestamp(float(bucket_start.value()))
                        for l in read_lines(r):
                            try:
                                if not l.endswith(b'\n'):
                                    raise Exception(f'{l!r} does not end in \\n')
                                time_delta, col_values, size=decode_timestamped_record(l,self.schema)
                            except Exception as e:
                                corruption_handler(e)
                            else:
                                timestamp=ts+time_delta
                                if timestamp >= begin and timestamp < end:
                                    yield (timestamp, col_values)
                                    records_yielded += 1
                                    bytes_yielded += size
                                    if records_yielded==max_records or bytes_yielded>=max_bytes:
                                        return
                                    pass
                                pass
                            pass
                        pass
                    pass
                except Exception as e:
                    corruption_handler(e)
                pass
            pass
        except Exception:
            raise in_function_context(PerfLog.fetch,vars()) from None
        pass

    def get_some_unseen_data(
            self,
            seen:Dict[Tuple[BucketStart,BucketID], ByteCount],
            max_bytes:ByteCount,
            read_failed:Callable[[Exception],Any]
    ) -> Dict[Tuple[BucketStart,BucketID], Tuple[FilePosition,bytes]]:
        '''return up to {max_bytes} bytes of unseen data from {self} having seen {seen}
           - new data might end with a partial record i.e. not record-aware
           - new data might truncate seen data
           - new data size 0 means bucket does not exist (any longer)
           - read failures are passed to read_failed, skipping that file if read_failed returns
        '''
        try:
            result={}
            result_size=ByteCount(0)
            for bucket, size in self.__tstore.list_unseen(seen).items():
                if size==ByteCount(0):
                    result[bucket]=(FilePosition(0),b'')
                    continue
                bucket_start, bucket_id=bucket
                seen_size=seen.get(bucket,ByteCount(0))
                if seen_size>size:
                    seen_size=ByteCount(0)
                    pass
                r:Reader
                try:
                    with self.__tstore.new_reader(bucket_start,bucket_id) as r:
                        read_size=size-seen_size
                        if read_size>max_bytes-result_size:
                            read_size=max_bytes-result_size
                            pass
                        r.seek_to(FilePosition(seen_size.value()))
                        data=r.read(read_size)
                        result[bucket]=(FilePosition(seen_size.value()),data)
                        result_size+=ByteCount(len(data))
                        pass
                except Exception as e:
                    read_failed(e)
                    pass
                if result_size==max_bytes:
                    break
                pass
            return result
        except Exception as e:
            raise in_function_context(PerfLog.get_some_unseen_data,vars()) from None
        pass

    def new_recorder(self):
        return Recorder(self.storage_path,self.schema,self.__tstore)

    def new_tracker(self):
        return Tracker(self.storage_path,self.__tstore)

    pass

def validate_record(record:List, schema:Dict[ColName,ColType]) -> List:
    '''validate record {record} against schema {schema}
       - returns record'''
    try:
        if len(record) != len(schema):
            raise Exception(f'schema expects {len(schema)} values but record has {len(record)}')
        for value,s in zip(record, schema.items()):
            col_name, col_type = s
            try:
                validate_col(value,col_type)
            except Exception:
                raise in_context(f'validate {col_name} value')
            pass
        return record
    except Exception as e:
        raise in_function_context(validate_record,vars()) from None
    pass

def encode_timestamped_record(time_delta:Duration, record:List, schema:Dict[ColName,ColType]) -> bytes:
    '''encode record {record} and time delta {time_delta} assuming record conforms to schema {schema}'''
    try:
        assert time_delta>=Duration(0)
        return (json.dumps([time_delta.value()]+validate_record(record, schema))+'\n').encode('utf-8')
    except:
        raise in_function_context(encode_timestamped_record,vars()) from None
    pass

def decode_timestamped_record(data:bytes, schema:Dict[ColName,ColType]) -> Tuple[
        Duration, #time_delta
        List[Union[str,int,float,bool,None,List,Dict]], #record
        ByteCount]: # len(data)
    '''decode time delta and record from {data!r} assuming record conforms to schema {schema}'''
    try:
        t:type
        Assert(data).endswith(b'\n')
        x=json.loads(data.decode('utf-8')[:-1])
        record:List[Union[str,int,float,bool,None,List,Dict]]=[]
        if isinstance(x, list):
            if isinstance(x[0], float):
                time_delta=Duration(x[0])
                assert time_delta>=Duration(0), time_delta
                values=x[1:]
                for i, col_type_spec in enumerate(schema.values()):
                    record.append(validate_col(values[i],col_type_spec))
                    pass
                return time_delta, record, ByteCount(len(data))
            else:
                t=type(x[0])
                raise Exception(f'time delta {x[0]!r} is of type {t}, which is not a float')
            pass
        else:
            t=type(x)
            raise Exception(f'{x!r} is of type {t}, which is not a list')
            pass
    except Exception as e:
        raise in_function_context(decode_timestamped_record,vars()) from None
    pass

    
def read_lines(r:Reader) -> Generator[bytes,None,None]:
    r'''read all \n-separated lines from PerLog reader {r}
        - yield one record at a time, including any terminating "\n"
          * where the file does not end in \n the last record yielded will have no "\n"'''
    try:
        buffered=io.BytesIO()
        next:bytes=r.read(ByteCount(1024))
        while len(next):
            recs=next.split(b'\n')
            while len(recs)>1:
                buffered.write(recs[0])
                buffered.write(b'\n')
                yield buffered.getvalue()
                buffered=io.BytesIO()
                recs=recs[1:]
                pass
            buffered.write(recs[0])
            next=r.read(ByteCount(1024))
            pass
        rem=buffered.getvalue()
        if len(rem):
            yield rem
        pass
    except Exception as e:
        raise in_function_context(read_lines,vars()) from None
    pass
                
                
attrs_schema=jsonschema.Schema({
    'schema':[ [str,OneOf('str','int','float','bool','(str)','(int)','(float)','(bool)','json')] ]
})

def validate_str(value)->str:
    if isinstance(value,str):
        return value
    t=type(value)
    raise Exception(f'{value!r} (of type {t}) is not a str')

def validate_int(value)->int:
    if isinstance(value,int):
        return value
    t=type(value)
    raise Exception(f'{value!r} (of type {t}) is not a int')

def validate_float(value)->float:
    if isinstance(value,float):
        return value
    if isinstance(value,int):
        return value
    t=type(value)
    raise Exception(f'{value!r} (of type {t}) is not a float')

def validate_bool(value)->bool:
    if isinstance(value,bool):
        return value
    t=type(value)
    raise Exception(f'{value!r} (of type {t}) is not a bool')

def validate_optional_str(value)->Optional[str]:
    if value is None:
        return value
    return validate_str(value)

def validate_optional_int(value)->Optional[int]:
    if value is None:
        return value
    return validate_int(value)

def validate_optional_float(value)->Optional[float]:
    if value is None:
        return value
    return validate_float(value)

def validate_optional_bool(value)->Optional[bool]:
    if value is None:
        return value
    return validate_bool(value)

def validate_col(value:Any,col_type:ColType)->Union[str,int,float,None,bool,List,Dict]:
    '''verify that value of type value.__class__ is valid for column type {col_type}'''
    try:
        if col_type=='str':
            return validate_str(value)
        if col_type=='int':
            return validate_int(value)
        if col_type=='float':
            return validate_float(value)
        if col_type=='bool':
            return validate_bool(value)
        if col_type=='(str)':
            return validate_optional_str(value)
        if col_type=='(int)':
            return validate_optional_int(value)
        if col_type=='(float)':
            return validate_optional_float(value)
        if col_type=='(bool)':
            return validate_optional_bool(value)
        if col_type=='json':
            # assume caller will effectively validate
            return cast(Union[str,int,float,None,bool,List,Dict],value)
        assert False, f'unknown col type {col_type}'
    except Exception as e:
        raise in_function_context(validate_col,vars())
    pass

PERFLOG_ATTRS='perflog.json'

def read_attrs(storage_path:Path,
               attrs_file=PERFLOG_ATTRS) -> Dict[ColName,ColType]:  # schema
    '''read PerfLog attrs from {storage_path}/{attrs_file}'''
    try:
        with FileReader(storage_path / attrs_file) as f:
            attrs=attrs_schema.validate(json.loads(f.read(f.size())))
            Assert(attrs).isInstanceOf(dict)
            schema={ColName(col_name): cast(ColType,col_type)
                    for col_name,col_type in cast(List[Tuple[str,str]],attrs['schema'])}
            pass
        return schema
    except Exception as e:
        raise in_function_context(read_attrs,vars()) from None
    pass

def write_attrs(storage_path:Path,
                schema:Dict[ColName,ColType],
                file_creation_mode:FileMode,
                attrs_file=PERFLOG_ATTRS) -> ByteCount:
    '''write PerfLog attrs schema {schema} to ''' \
    '''{storage_path}/{attrs_file}
       - returns number of bytes written'''
    try:
        col_name:ColName
        col_type:ColType
        x=attrs_schema.validate({
            'schema':[ [col_name.value(),col_type] for col_name, col_type in schema.items()]
        })
        y=json.dumps(x).encode('utf-8')
        
        with FileWriter(storage_path / attrs_file,
                        mode=file_creation_mode-FileMode(0o111),
                        must_not_exist=True) as f:
            f.output.write(y)
            pass
        return ByteCount(len(y))
    except Exception as e:
        raise in_function_context(write_attrs,vars()) from None
    pass

def trim_trailing_partial_record(
        data:Dict[Tuple[BucketStart,BucketID], Tuple[FilePosition,bytes]])->None:
    '''trim data so it ends with a complete record
       - note you can use this to ensure Tracker always writes complete records'''
    if not len(data):
        return
    last_bucket=list(data.keys())[-1]
    position,d=data[last_bucket]
    trimmed=d[0:d.rfind(b'\n')+1]
    if len(trimmed)==0:
        del data[last_bucket]
    else:
        data[last_bucket]=(position,trimmed)
        pass
    pass



# coding: utf-8

# Copyright (c) 2018 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
import traceback
import sys
import string
from typing import Sequence,Callable,Literal,Dict,List,Tuple,Any,Type
from types import TracebackType

class FileAndLine(object):
    def __init__(self,file:str|None=None,line:int|None=None,readable:bool=True)->None:
        self.file=file
        self.line=line
        self.readable=readable
        pass
    def setTo(self,file:str,line:int)->None:
        self.file=file
        self.line=line
        pass
    def __str__(self)->str:
        if self.file:
            return '{file}:{line}: '.format(**self.__dict__)
        return ''
    pass

class Xn:
    """Capture cause and context.
    """
    def __init__(self, cause:object)->None:
        '''cause is convertable to a string using str'''
        ''' e.g. cause can be an exception or a string like "file not found"'''
        '''cause can also have a xju_xn_readable_repr() method, in which case that'''
        '''will be used by xju_xn_readable_repr() below'''
        self.cause = (cause,FileAndLine()) # file,line set below
        self.context:List[Tuple[str,FileAndLine]] = [] # (text,FileAndLine)
        pass

    def __str__(self)->str:
        '''programmer friendly format, each context and cause includes
        file and line, and intermediate stack entries are included
        - see unit test below for example'''
        result = ''
        x = ''.join([
            '{fl}failed to {s} because\n'.format(**vars())
            for s,fl in reversed(self.context)])
        m = str(self.cause[0]) or type(self.cause[0]).__name__
        y = f'{self.cause[1]}{m}'
        return x+y

    def xju_xn_readable_repr(self)->str:
        '''human (non-programmer) readable representation, omitting file
        and line, omitting intermediate stack entries, and producing a
        proper sentence i.e. capitalised and ending in full stop
        - see unit test below for example'''
        result = ''
        x:str = ''.join([
            'failed to {s} because\n'.format(**vars())
            for s,fl in reversed(self.context)
            if fl.readable])
        if hasattr(self.cause[0], 'xju_xn_readable_repr'):
            y:str=self.cause[0].xju_xn_readable_repr()
        else:
            y = str(self.cause[0]) or type(self.cause[0]).__name__
            pass
        return capitalise(x+y+'.')
    pass

def readable_repr(e:BaseException)->str:
    r=getattr(e,'xju_xn_readable_repr',None)
    if callable(r):
        return str(r())
    return str(e)

def capitalise(s:str)->str:
    if s and s[0]!=s[0].upper():
        return s[0].upper()+s[1:]
    return s

def in_function_context(function:Callable,
                        vars:Dict[str,Any]={},
                        exceptionInfo:None|tuple[Type[BaseException],BaseException,TracebackType]=None,
                        fl:None|tuple[str,int]=None)->BaseException:
    """Make a Xn that includes exception info and context as first_line_of(f.__doc__).format(**vars()).

       - if exceptionInfo[1] is already a Xn just add context
       - otherwise use exceptionInfo as cause for a new Xn
       - exceptionInfo defaults to sys.exc_info()
    """
    return in_context(first_line_of(function.__doc__).format(**vars),
                      exceptionInfo=exceptionInfo,
                      fl=fl)

def in_context(context:str,
               exceptionInfo:None|tuple[Type[BaseException],BaseException,TracebackType]=None,
               fl:None|tuple[str,int]=None)->BaseException:
    """Make a Xn that includes exception info and context.

       - if exceptionInfo[1] is already a Xn just add context
       - otherwise use exceptionInfo as cause for a new Xn
       - exceptionInfo defaults to sys.exc_info()
    """
    if exceptionInfo is None:
        t, e, b = sys.exc_info()
        assert t is not None
        assert e is not None
        assert b is not None
        exceptionInfo=(t,e,b)
    exceptionType,r,traceBack=exceptionInfo

    if not isinstance(r,Xn):
        #build new exception type derived from both original and Xn
        name=exceptionType.__name__
        def init(self,v):
            Xn.__init__(self,v)
            for a in set.difference(set(dir(v)),set(dir(Xn))):
                try:
                    setattr(self,a,getattr(v,a))
                except AttributeError:
                    pass
                pass
            pass
        def str_(self)->str:
            return Xn.__str__(self)
        def xju_xn_readable_repr(self)->str:
            return Xn.xju_xn_readable_repr(self)
        r=type(name,
               (Xn,exceptionType),
               {
                   '__init__':init,
                   '__str__':str_,
                   'xju_xn_readable_repr':xju_xn_readable_repr
               })(r)
    
    st=[tuple(_) for _ in traceback.extract_tb(traceBack)]
    # fill in most recent file,line (latest context or cause if no context)
    f,l=st[-1][0:2]
    assert isinstance(r, Xn)
    if r.context:
        if not r.context[-1][1].file:
            r.context[-1][1].file=f
            pass
        if not r.context[-1][1].line:
            r.context[-1][1].file=l
            pass
    else:
        r.cause[1].setTo(f,l)
        pass
    # add context entries for any in-between stack entries
    newContext=[(text,FileAndLine(file,line,False))
                for file,line,fname,text in reversed(st[0:-1])]
    r.context.extend(newContext)
    # add the supplied context, with unknown file and line
    f2=fl[0] if fl else None
    l2=fl[1] if fl else None
    r.context.append( (context,FileAndLine(f2,l2)) )
    traceBack.tb_next=None
    return r


def first_line_of(x:Any)->str:
    '''return first non-empty line of str({x}) stripped of leading and trailing whitespace'''
    return str(x).strip().split('\n')[0].strip()

def first_para_of(x:Any)->str:
    '''return first paragraph of str({x}) stripped of leading and trailing whitespace
    with separate lines stripped and joined by a single space

    - paragraph ends at blank line'''
    lines=[_.strip() for _ in str(x).strip().split('\n')]+['']
    end=lines.index('')
    return ' '.join(lines[0:end])

def desentence(s:str)->str:
    '''remove any trailing '.' and down-case first characters of {s}'''
    if s.endswith('.'): s=s[:-1]
    return s[0:1].lower()+s[1:]

def indent(prefix:str,s:str)->str:
    '''prefix all but first line of s by specified prefix'''
    return s.replace('\n','\n'+prefix)

class AllFailed(Exception):
    def __init__(self,causes:Sequence[BaseException])->None:
        self.causes=causes
        pass
    def __str__(self)->str:
        return ', and\n'.join([str(cause) for cause in self.causes])
    def xju_xn_readable_repr(self)->str:
        return '; and\n'.join(['- '+
                               indent('  ',desentence(readable_repr(cause)))
                               for cause in self.causes])
    pass




# coding: utf-8

# Copyright (c) 2018 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
from xju.xn import in_function_context,in_context
from xju import assert_

CTLs=set.union(set([chr(n) for n in range(0,32)]),set([chr(127)]))
separators=set('()<>@,;:\\"{} \t')

def validateToken(name):
    '''validate RFC2616 token {name!r}'''
    try:
        if not len(name): raise Exception(f'{name!r} is empty')
        for i,c in enumerate(name):
            o=ord(c)
            try:
                if ord(c)<0 or ord(c)>127:
                    raise Exception(f'{o} not 0..127')
                if c in CTLs:
                    raise Exception(f'{c!r} is a control character')
                if c in separators:
                    raise Exception(f'{c!r} is a separator')
            except:
                rest=name[i:]
                raise in_context(f'validate first char' if i==0 else f'validate char at ...{rest!r}') from None
            pass
        return name
    except:
        raise in_function_context(validateToken,vars()) from None
    pass




# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# CMC - Context Management, I don't know what the extra C stands for
#
from __future__ import annotations

import asyncio
import sys
import contextlib
from dataclasses import dataclass
from typing import TypeVar, Iterable, Dict as _Dict, overload, Tuple, Sequence, Union, Optional
from typing import ItemsView, KeysView
from typing import _GenericAlias # type: ignore  # mypy 1.0.0 :-(
from types import GenericAlias
from typing import Mapping, Type, List, Generic, Any, Callable, Awaitable, Self
from collections import OrderedDict
from collections.abc import Coroutine
import builtins
import threading
from xju.time import Duration
from xju.xn import in_function_context, in_context
from xju.assert_ import Assert

T = TypeVar('T')
U = TypeVar('U')

T_CMT = TypeVar('T_CMT', bound=Type[contextlib.AbstractContextManager])

def cmclass(cls:T_CMT) -> T_CMT:
    '''Class decorator that adds context management __enter__ and __exit__
       that enter and exit all type-hinted attributes implementing contextlib.AbstractContextManager
       then any existing __enter__.
       Note that to satisfy mypy, the decorated class must already
       implement contextlib.AbstractContextManager and the preferred way to do that is
       to inherit from xju.cmc.CM - see cmc.py.test for examples.
       See also async_cmclass, which provides an asyncio equivalent.
    '''
    base_classes_to_enter = [ base_class for base_class in cls.__bases__
                              if (
                                      base_class is not CM and
                                      issubclass(base_class, contextlib.AbstractContextManager)
                              ) ]
    attrs_to_enter = [ n for n, t in cls.__annotations__.items()
                         if _is_subclass(n, t, contextlib.AbstractContextManager)]
    # need a unique place to keep resources acquired by enter so exit can
    # exit them, only want resources for self, not subclasses (which do their own
    # handling). Note replacing . with _ could lead to clashes but there's no
    # way to avoid all clashes.
    resources_attr_name=f'{cls.__module__}.{cls.__name__}'.replace('.','_')
    orig_enter=cls.__dict__.get('__enter__',lambda self: self)
    def enter(self,
              base_classes_to_enter=base_classes_to_enter,
              attrs_to_enter=attrs_to_enter,
              resources_attr_name=resources_attr_name,
              orig_enter=orig_enter) -> Type[contextlib.AbstractContextManager[T]]:
        with contextlib.ExitStack() as tentative:
            for base_class in base_classes_to_enter:
                cm = _ClassCm(base_class, self)
                tentative.enter_context(cm)
                pass
            for n in attrs_to_enter:
                tentative.enter_context(getattr(self, n))
                pass
            assert getattr(self,resources_attr_name, None) is None, f'{cls.__module__}.{cls.__name__} has a f{resources_attr_name} attribute already. Perhaps it is inherited multiple times (which is not allowed by xju.cmc.cmclass decorator)?'
            orig_enter(self)
            setattr(self,resources_attr_name,tentative.pop_all())
        return self
    orig_exit=cls.__dict__.get('__exit__',lambda self,t,e,b: None)
    def exit(self,t,e,b,resources_attr_name=resources_attr_name) -> None:
        resources = getattr(self,resources_attr_name)
        setattr(self,resources_attr_name,None)
        with contextlib.closing(resources):
            return orig_exit(self,t,e,b)
    old_sa=cls.__setattr__
    def sa(self, name, value, attrs_to_enter=attrs_to_enter):
        if getattr(self,resources_attr_name,None) and name in attrs_to_enter:
            assert False, f'xju.cmc {self} has been entered so cannot replace context-managed attribute {cls.__module__}.{cls.__name__}.{name}'
        return old_sa(self,name,value)
    setattr(cls, '__enter__', enter)
    setattr(cls, '__exit__', exit)
    setattr(cls, '__setattr__', sa)
    cls.__annotations__['__enter__'] = type(enter)
    cls.__annotations__['__exit__'] = type(exit)
    return cls

class CM(contextlib.AbstractContextManager):
    def __enter__(self):
        assert False, f'{CM.__module__}.CM.__enter__ not overridden, have you forgotten @cmclass on {self.__module__}.{self.__class__.__name__}?'
        return self
    def __exit__(self, t, e, b) -> None:
        assert False, f'{CM.__module__}.CM.__exit__ not overridden, have you forgotten @cmclass on {self.__module__}.{self.__class__.__name__}?'
        pass

K = TypeVar('K')
V = TypeVar('V', bound=contextlib.AbstractContextManager)

class Dict(Mapping[K, V], contextlib.AbstractContextManager):
    '''dictionary that enters and exits its values, all of which must be context managers
       note that values may be added and removed before and/or after self is entered:
         before: the value will be entered when self is entered
         after:  the value will be entered before it is added (noting that any existing
                 value will be exited once (if) it is removed) - the enter-new-exit-old 
                 order might be important to you i.e. might not be what you need!'''
    entered = False
    x:_Dict[K,V]

    @overload
    def __init__(self):
        '''new empty Dict'''

    @overload
    def __init__(self, x:_Dict[K,V]):
        '''initialise with value from {x} assuming those values have not been "entered"'''

    @overload
    def __init__(self, x:Iterable[Tuple[K,V]]):
        '''initialise with value from {x} assuming those values have not been "entered"'''

    def __init__(self, *args, **kwargs):
        self.x = OrderedDict(*args, **kwargs)
        pass
    
    def __enter__(self):
        '''"enters" all values in order they were inserted'''
        with contextlib.ExitStack() as tentative:
            for k,v in self.x.items():
                tentative.enter_context(v)
            tentative.pop_all()
            pass
        self.entered = True
        return self

    def __exit__(self, t, e, b):
        '''"exits" all values in order they were inserted'''
        if self.entered:
            with contextlib.ExitStack() as resources:
                for k,v in self.x.items():
                    resources.push(v)
                    pass
                pass
            pass
        pass

    def __contains__(self, key:object) -> bool:
        return key in self.x

    def __delitem__(self, key:K):
        v = self.x[key]
        del self.x[key]
        if self.entered:
            with contextlib.ExitStack() as f:
                f.push(v)
                pass
            pass
        pass

    def __getitem__(self, key:K) -> V:
        return self.x[key]
    
    def __iter__(self):
        return iter(self.x.keys())

    def __len__(self) -> int:
        return len(self.x)

    def __repr__(self) -> str:
        return repr(self.x)

    def __setitem__(self, key:K, value:V) -> None:
        '''replace any current value of {key} with {value}'''
        '''- "enters" the new value and then "exits" any old value'''
        old = self.x.get(key,None)
        if old is not value:
            if self.entered:
                value.__enter__()
                if old is None:
                    self.x[key]=value
                else:
                    with contextlib.ExitStack() as f:
                        f.push(old)
                        self.x[key]=value
                    pass
                pass
            else:
                self.x[key]=value
            pass
        pass
    
    def __sizeof__(self) -> int:
        return self.x.__sizeof__()

    def clear(self):
        while len(self):
            self.popitem()
            pass
        pass

    def get(self, key, default=None):
        return self.x.get(key, default)

    def items(self) -> ItemsView[K,V]:
        '''items in order they were first inserted'''
        return self.x.items()

    def keys(self) -> KeysView[K]:
        '''keys in order they were first inserted'''
        return self.x.keys()

    @overload
    def pop(self, key:K) -> V:
        '''pop and "exit" value of {key} if self has a value for it
           otherwise raise KeyError'''

    @overload
    def pop(self, key:K, default:T) -> V|T:
        '''pop and "exit" value of {key} if self has a value for it
        otherwise return default'''

    def pop(self, key, default=None):
        if default is None:
            v = self.x.pop(key)
            if self.entered:
                with contextlib.ExitStack() as f:
                    f.push(v)
                pass
            return v
        else:
            try:
                v = self.x.pop(key)
                if self.entered:
                    with contextlib.ExitStack() as f:
                        f.push(v)
                    pass
                return v
            except KeyError:
                return default
            pass
        pass

    def popitem(self) -> Tuple[K, V]:
        '''pop and "exit" most recently added item'''
        k, v = self.x.popitem()
        if self.entered:
            with contextlib.ExitStack() as f:
                f.push(v)
                pass
            pass
        return k, v

    @overload
    def setdefault(self, key:K, default:V) -> V:
        '''set value of {key} to default "entered" if it has no value
           return value of {key}'''

    @overload
    def setdefault(self, key:K, default:None) -> Optional[V]:
        '''same as get(key,None)'''

    def setdefault(self, key, default):
        if default is not None:
            if not key in self:
                self[key] = default
            return self.x.get(key)
        else:
            return self.x.get(key, None)
        pass

    @overload
    def update(self, x:_Dict[K, V]):
        '''update self from {x} assuming its values are not yet entered'''

    @overload
    def update(self, x:Iterable[Tuple[K,V]]):
        '''update self from {x} assuming its values are not yet entered'''

    def update(self, x):
        if getattr(x,'keys',None):
            for k in x.keys():
                self[k] = x[k]
                pass
            pass
        else:
            for k,v in x:
                self[k] = v
                pass
            pass
        pass

class Opt(Generic[V],contextlib.AbstractContextManager[None|V]):
    '''holds an optional context manager
       - value can be set (see set()) and cleared (see clear()) at any time
       - if value is set while active, new value is entered and then old value if any is exited
       - if value is cleared while active, old value if any is exited
    '''
    entered = False
    x:None|V = None

    @overload
    def __init__(self):
        '''new empty optional'''
        pass
    @overload
    def __init__(self, x:Opt[V]):
        '''initialise with value from {x} assuming those values have not been "entered"'''
        pass
    @overload
    def __init__(self, x:V):
        '''initialise with value {x} assuming it has not been "entered"'''
        pass
    def __init__(self, *args):
        if len(args):
            match args[0]:
                case Opt() as y:
                    self.x = y.x
                    pass
                case v:
                    self.x=v
                    pass
            pass
        pass
    
    def __enter__(self) -> None|V:
        '''"enters" value if present'''
        result:None|V = None
        if self.x is not None:
            result=self.x.__enter__()
            pass
        self.entered=True
        return result

    def __exit__(self, t, e, b):
        '''"exits" value if present'''
        assert self.entered, self.x
        if self.x:
            with contextlib.ExitStack() as resources:
                resources.push(self.x)
                pass
            pass
        self.entered=False
        pass

    def __repr__(self) -> str:
        return f"Opt({self.x})"

    @overload
    def set(self,value:V) -> None|V:
        '''replace any current value of {key} with {value}
           - if active, "enters" the new value and then "exits" any old value
           - returns old value if any
        '''
        pass
    @overload
    def set(self,value:Opt[V]) -> None|V:
        '''replace any current value of {key} with {value}
           - if active, "enters" any new value and then "exits" any old value
           - returns old value if any
           - noop if value is the old value, e.g. x.set(x.get())
        '''
        pass
    def set(self,value) -> None|V:
        '''replace any current value of {key} with {value}
           - if active, "enters" any new value and then "exits" any old value
           - returns old value if any
           - noop if value is the old value, e.g. x.set(x.get())
        '''
        if isinstance(value,Opt):
            if value.x is None:
                return self.clear()
            else:
                return self.set_value(value.x)
        return self.set_value(value)

    def set_value(self,value:V) -> None|V:
        '''replace any current value with {value}
           - assumes value has not been entered
           - if active, "enters" new value and then "exits" any old value
           - noop if value is the old value, e.g. x.set(x.get())
           - returns old value if any
        '''
        old = self.x
        if old is not value:
            if self.entered:
                value.__enter__()
                if old is None:
                    self.x=value
                else:
                    with contextlib.ExitStack() as f:
                        f.push(old)
                        self.x=value
                    pass
                pass
            else:
                self.x=value
            pass
        return old
    
    def clear(self) -> None|V:
        '''clear any current value
           - if active, "exits" old value
           - returns old value
        '''
        old = self.x
        if old is not None:
            if self.entered:
                with contextlib.ExitStack() as f:
                    f.push(old)
                    self.x=None
                pass
            else:
                self.x=None
            pass
        return old

    def get(self) -> None|V:
        '''get current value if any'''
        return self.x

class _ClassCm(contextlib.AbstractContextManager):
    '''target object {x} subclass {cls} context management methods
       - so for x: X where X(Y), __ClassCM(Y, x) will call Y.__enter__
         and Y.__exit__ directly avoiding any X.__enter__/__exit__
         overrides'''
    cls:Type
    x:object

    def __init__(self, cls, x):
        self.cls=cls
        self.x=x
        pass
    def __enter__(self):
        self.cls.__enter__(self.x)
        return self
    def __exit__(self, t, e, b):
        return self.cls.__exit__(self.x, t, e, b)
    pass


def _is_subclass(n:str, t1, t2:type):
    '''check whether {n}'s type {t1} is a sub-class of {t2}'''
    try:
        Assert(t2).isInstanceOf(type)
        if issubclass(type(t1),GenericAlias):
            return _is_subclass(n,t1.__origin__,t2)
        elif issubclass(type(t1),_GenericAlias):
            return _is_subclass(n,t1.__origin__,t2)
        elif isinstance(t1,type):
            return issubclass(t1, t2)
        else:
            return False
        pass
    except Exception:
        raise in_function_context(_is_subclass,vars()) from None
    pass

ResultType=TypeVar('ResultType')

# deprecated since xju 1.2.13, use AsyncTask, because...
# this task mistakenly takes a coroutine, which means that if the task is never
# run, the program warns of coroutine-not-awaited at exit
class Task(contextlib.AbstractAsyncContextManager,Generic[ResultType]):
    '''asyncio Task context manager that guarentees Task awaited before exit
       - note that normally, application will have waited for the task
         within the context; cancel-on-exit ensures clean up on exception
         e.g. when using gather() with its default return_exceptions=False'''
    coroutine: Coroutine[Any,Any,ResultType]
    task: None | asyncio.Task[ResultType]
    def __init__(self, coroutine: Coroutine[Any,Any,ResultType]) -> None:
        self.coroutine=coroutine
        self.task=None
    async def __aenter__(self) -> asyncio.Task[ResultType]:
        '''create and return task'''
        assert self.task is None
        self.task=asyncio.create_task(self.coroutine)
        return self.task
    async def __aexit__(self, t, e, b) -> None:
        '''cancel and await task'''
        assert self.task is not None
        task=self.task
        self.task=None
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass

class Thread(Generic[ResultType],contextlib.AbstractContextManager[Callable[[],ResultType]]):
    __t: threading.Thread
    __result: ResultType | Exception
    
    def __init__(self, run: Callable[[],ResultType], stop: Callable[[],None]):
        '''at context entry call {run}() in new thread
           at context exit call {stop}() and wait for thread to exit
           result is not available until context exited'''
        self.run = run
        self.stop = stop
        self.__t = threading.Thread(target=self.__run)
        self.__result = Exception('thread not run')
        pass

    def __str__(self):
        return f'thread {self.__t} to run {self.run}'

    def __enter__(self) -> Callable[[],ResultType]:
        '''start {self}'''
        try:
            self.__t.start()
            return self.result
        except Exception:
            raise in_function_context(Thread.__enter__,vars())
        pass

    def __exit__(self, t, e, b):
        '''stop (via {self.stop}()) and wait for (i.e. join) {self}'''
        try:
            self.stop()
            self.__t.join()
        except Exception:
            raise in_function_context(Thread.__exit__, vars())
        pass

    def result(self) -> ResultType:
        '''assuming context has exited, return/raise result of run()'''
        if isinstance(self.__result, Exception):
            raise self.__result
        return self.__result

    def __run(self) -> None:
        try:
            self.__result = self.run()
        except Exception as e:
            self.__result = e
            pass
    pass

class Mutex:
    def __init__(self):
        self.m = threading.Lock()
        pass
    pass

class Lock(contextlib.AbstractContextManager):
    def __init__(self, m: Mutex):
        self.m = m
        self.active=False
        pass

    def __enter__(self) -> Self:
        self.m.m.acquire()
        self.active=True
        return self

    def __exit__(self,t,e,b) -> None:
        self.active=False
        self.m.m.release()
        pass

    pass

class Condition:
    def __init__(self, m:Mutex):
        self.m = m
        self.c = threading.Condition(m.m)

    def wait_for(self, l:Lock, timeout:Duration) -> None:
        assert l.m == self.m and l.active
        self.c.wait(timeout=timeout.value())
        pass

    def notify_all(self, l:Lock) -> None:
        assert l.m == self.m and l.active
        self.c.notify_all()
    pass

ACMT = TypeVar('ACMT', bound=contextlib.AbstractAsyncContextManager)
T_ACMT = TypeVar('T_ACMT', bound=Type[contextlib.AbstractAsyncContextManager])

def async_cmclass(cls:T_ACMT) -> T_ACMT:
    '''Class decorator that adds async context management __aenter__ and __aexit__
       that enter and exit all type-hinted attributes implementing 
       contextlib.AbstractAsyncContextManager and contextlib.AbstractContextManager
       then any existing __aenter__.
       Note that to satisfy mypy, the decorated class must already
       implement contextlib.AbstractAsyncContextManager and the preferred way to do that is
       to inherit from xju.cmc.AsyncCM - see cmc.py.test for examples.
    '''
    base_classes_to_enter = _make_base_classes_to_enter(cls) # type: ignore  # mypy 1.4.1 gets confused
    attrs_to_enter = _make_attrs_to_enter(cls) # type: ignore  # mypy 1.4.1 gets confused
    # need a unique place to keep resources acquired by enter so exit can
    # exit them, only want resources for self, not subclasses (which do their own
    # handling). Note replacing . with _ could lead to clashes but there's no
    # way to avoid all clashes.
    resources_attr_name=f'xju_cmc_async_cmclass_resources_{cls.__module__}.{cls.__name__}'.replace('.','_')
    orig_enter=cls.__dict__.get('__aenter__')
    async def aenter(self,
                     base_classes_to_enter=base_classes_to_enter,
                     attrs_to_enter=attrs_to_enter,
                     resources_attr_name=resources_attr_name,
                     orig_enter=orig_enter) -> Type[contextlib.AbstractAsyncContextManager[T]]:
        async with contextlib.AsyncExitStack() as tentative:
            for f in base_classes_to_enter:
                cm:contextlib.AbstractAsyncContextManager = f(self)
                await tentative.enter_async_context(cm)
                pass
            for f in attrs_to_enter.values():
                cm = f(self)
                await tentative.enter_async_context(cm)
                pass
            assert getattr(self,resources_attr_name, None) is None, f'{cls.__module__}.{cls.__name__} has a f{resources_attr_name} attribute already. Perhaps it is inherited multiple times (which is not allowed by xju.cmc.async_cmclass decorator)?'
            if orig_enter is not None:
                await orig_enter(self)
                pass
            setattr(self,resources_attr_name,tentative.pop_all())
        return self
    orig_exit=cls.__dict__.get('__aexit__',None)
    async def aexit(self,t,e,b,resources_attr_name=resources_attr_name) -> None:
        resources = getattr(self,resources_attr_name)
        setattr(self,resources_attr_name,None)
        async with contextlib.aclosing(resources):
            if orig_exit is not None:
                await orig_exit(self,t,e,b)
                pass
            pass
        pass
    old_sa=cls.__setattr__
    def sa(self, name, value, attrs_to_enter=attrs_to_enter):
        if getattr(self,resources_attr_name,None) and name in attrs_to_enter:
            raise Exception(
                f'xju.cmc {self} has been entered so cannot replace context-managed attribute {cls.__module__}.{cls.__name__}.{name}')
        return old_sa(self,name,value)
    setattr(cls, '__aenter__', aenter)
    setattr(cls, '__aexit__', aexit)
    setattr(cls, '__setattr__', sa)
    cls.__annotations__['__aenter__'] = type(aenter)
    cls.__annotations__['__aexit__'] = type(aexit)
    return cls

class AsyncCM(contextlib.AbstractAsyncContextManager):
    '''base class for use with async_cm_class decorator above, to appease type checker'''
    async def __aenter__(self) -> Self:
        assert False, f'{CM.__module__}.AsyncCM.__aenter__ not overridden, have you forgotten @async_cmclass on {self.__module__}.{self.__class__.__name__}?'
        return self
    async def __aexit__(self, t, e, b) -> None:
        assert False, f'{CM.__module__}.AsyncCM.__aexit__ not overridden, have you forgotten @async_cmclass on {self.__module__}.{self.__class__.__name__}?'
        pass

AsyncV = TypeVar('AsyncV', bound=contextlib.AbstractAsyncContextManager)

class AsyncDict(Mapping[K, AsyncV], contextlib.AbstractAsyncContextManager):
    '''dictionary that enters and exits its values, all of which must be context managers
       note that values may be added and removed before and/or after self is entered:
         before: the value will be entered when self is entered
         after:  the value will be entered before it is added (noting that any existing
                 value will be exited once (if) it is removed) - the enter-new-exit-old 
                 order might be important to you i.e. might not be what you need!
       note that x['a']==y is not implementable, use x.set('a',y)
       note that del x['a'] is not implementable, use x.pop('a')
    '''
    entered = False
    x:_Dict[K,AsyncV]

    @overload
    def __init__(self):
        '''new empty Dict'''

    @overload
    def __init__(self, x:_Dict[K,AsyncV]):
        '''initialise with value from {x} assuming those values have not been "entered"'''

    @overload
    def __init__(self, x:Iterable[Tuple[K,AsyncV]]):
        '''initialise with value from {x} assuming those values have not been "entered"'''

    def __init__(self, *args, **kwargs):
        self.x = OrderedDict(*args, **kwargs)
        pass
    
    async def __aenter__(self):
        '''"enters" all values in order they were inserted'''
        async with contextlib.AsyncExitStack() as tentative:
            for k,v in self.x.items():
                await tentative.enter_async_context(v)
            tentative.pop_all()
            pass
        self.entered = True
        return self

    async def __aexit__(self, t, e, b):
        '''"exits" all values in order they were inserted'''
        if self.entered:
            async with contextlib.AsyncExitStack() as resources:
                for k,v in reversed(self.x.items()):
                    resources.push_async_exit(v)
                    pass
                pass
            pass
        pass

    def __contains__(self, key:object) -> bool:
        return key in self.x

    def __getitem__(self, key:K) -> AsyncV:
        return self.x[key]
    
    def __iter__(self):
        return iter(self.x.keys())

    def __len__(self) -> int:
        return len(self.x)

    def __repr__(self) -> str:
        return repr(self.x)

    async def set(self, key:K, value:AsyncV) -> None:
        '''replace any current value of {key} with {value}'''
        '''- "enters" the new value and then "exits" any old value'''
        old = self.x.get(key,None)
        if old is not value:
            if self.entered:
                await value.__aenter__()
                if old is None:
                    self.x[key]=value
                else:
                    async with contextlib.AsyncExitStack() as f:
                        f.push_async_exit(old)
                        self.x[key]=value
                    pass
                pass
            else:
                self.x[key]=value
            pass
        pass
    
    def __sizeof__(self) -> int:
        return self.x.__sizeof__()

    async def clear(self):
        while len(self):
            await self.popitem()
            pass
        pass

    def get(self, key, default=None):
        return self.x.get(key, default)

    def items(self) -> ItemsView[K,AsyncV]:
        '''items in order they were first inserted'''
        return self.x.items()

    def keys(self) -> KeysView[K]:
        '''keys in order they were first inserted'''
        return self.x.keys()

    @overload
    async def pop(self, key:K) -> AsyncV:
        '''pop and "exit" value of {key} if self has a value for it
           otherwise raise KeyError'''

    @overload
    async def pop(self, key:K, default:T) -> AsyncV|T:
        '''pop and "exit" value of {key} if self has a value for it
        otherwise return default'''

    async def pop(self, key, default=None):
        if default is None:
            v = self.x.pop(key)
            if self.entered:
                async with contextlib.AsyncExitStack() as f:
                    f.push_async_exit(v)
                pass
            return v
        else:
            try:
                v = self.x.pop(key)
                if self.entered:
                    async with contextlib.AsyncExitStack() as f:
                        f.push_async_exit(v)
                    pass
                return v
            except KeyError:
                return default
            pass
        pass

    async def popitem(self) -> Tuple[K, AsyncV]:
        '''pop and "exit" most recently added item'''
        k, v = self.x.popitem()
        if self.entered:
            async with contextlib.AsyncExitStack() as f:
                f.push_async_exit(v)
                pass
            pass
        return k, v

    @overload
    async def setdefault(self, key:K, default:AsyncV) -> AsyncV:
        '''set value of {key} to default "entered" if it has no value
           return value of {key}'''

    @overload
    async def setdefault(self, key:K, default:None) -> AsyncV|None:
        '''same as get(key,None)'''

    async def setdefault(self, key, default):
        if default is not None:
            if not key in self:
                await self.set(key,default)
            return self.x.get(key)
        else:
            return self.x.get(key, None)
        pass

    @overload
    async def update(self, x:_Dict[K, AsyncV]):
        '''update self from {x} assuming its values are not yet entered
           - note if k,v from x is alredy in self, v is not re-entered
        '''

    @overload
    async def update(self, x:Iterable[Tuple[K,AsyncV]]):
        '''update self from {x} assuming its values are not yet entered
           - note if k,v from x is alredy in self, v is not re-entered
        '''
    async def update(self, x):
        if getattr(x,'keys',None):
            for k in x.keys():
                await self.set(k,x[k])
                pass
            pass
        else:
            for k,v in x:
                await self.set(k,v)
                pass
            pass
        pass

class AsyncOpt(Generic[AsyncV],contextlib.AbstractAsyncContextManager[None|AsyncV]):
    '''holds an optional context manager
       - value can be set (see set()) and cleared (see clear()) at any time
       - if value is set while active, new value is entered and then old value if any is exited
       - if value is cleared while active, old value if any is exited
    '''
    entered = False
    x:None|AsyncV = None

    @overload
    def __init__(self):
        '''new empty optional'''
        pass
    @overload
    def __init__(self, x:AsyncOpt[AsyncV]):
        '''initialise with value from {x} assuming those values have not been "entered"'''
        pass
    @overload
    def __init__(self, x:AsyncV):
        '''initialise with value {x} assuming it has not been "entered"'''
        pass
    def __init__(self, *args, **kwargs):
        if len(args):
            match args[0]:
                case AsyncOpt() as y:
                    self.x = y.x
                    pass
                case v:
                    self.x=v
                    pass
            pass
        pass
    
    async def __aenter__(self) -> None|AsyncV:
        '''"enters" value if present'''
        result:None|AsyncV = None
        if self.x is not None:
            result=await self.x.__aenter__()
            pass
        self.entered=True
        return result

    async def __aexit__(self, t, e, b):
        '''"exits" value if present'''
        assert self.entered, self.x
        if self.x:
            async with contextlib.AsyncExitStack() as resources:
                resources.push_async_exit(self.x)
                pass
            pass
        self.entered=False
        pass

    def __repr__(self) -> str:
        return f"AsyncOpt({self.x})"

    @overload
    async def set(self,value:AsyncV) -> None|AsyncV:
        '''replace any current value with {value}
           - "enters" the new value and then "exits" any old value
           - assumes value has not been entered
           - if active, "enters" any new value and then "exits" any old value
           - returns old value
           - noop if value is the old value, e.g. x.set(x.get())
        '''
        pass
    @overload
    async def set(self,value:AsyncOpt[AsyncV]) -> None|AsyncV:
        '''replace any current value with {value}
           - assumes value has not been entered
           - if active, "enters" any new value and then "exits" any old value
           - returns old value
           - noop if value is the old value, e.g. x.set(x.get())
        '''
        pass
    async def set(self,value) -> None|AsyncV:
        if isinstance(value,AsyncOpt):
            if value.x is None:
                return await self.clear()
            else:
                return await self.set_value(value.x)
        return await self.set_value(value)

    async def set_value(self,value:AsyncV) -> None|AsyncV:
        '''replace any current value with {value}
           - assumes value has not been entered
           - if active, "enters" new value and then "exits" any old value
           - returns old value
           - noop if value is the old value, e.g. x.set(x.get())
        '''
        old = self.x
        if old is not value:
            if self.entered:
                await value.__aenter__()
                if old is None:
                    self.x=value
                else:
                    async with contextlib.AsyncExitStack() as f:
                        f.push_async_exit(old)
                        self.x=value
                    pass
                pass
            else:
                self.x=value
            pass
        return old

    async def clear(self) -> None|AsyncV:
        '''clear any current value
           - if active, "exits" old value
           - returns old value
        '''
        old = self.x
        if old is not None:
            if self.entered:
                async with contextlib.AsyncExitStack() as f:
                    f.push_async_exit(old)
                    self.x=None
                pass
            else:
                self.x=None
            pass
        return old

    def get(self) -> None|AsyncV:
        '''get current value if any'''
        return self.x

class AsyncTask(contextlib.AbstractAsyncContextManager,Generic[ResultType]):
    '''asyncio Task context manager that guarentees Task awaited before exit
       - note that normally, application will have waited for the task
         within the context; cancel-on-exit ensures clean up on exception
         e.g. when using gather() with its default return_exceptions=False'''
    function: Callable[[],Coroutine[Any,Any,ResultType]]
    task: None | asyncio.Task[ResultType]
    def __init__(self, function: Callable[[],Coroutine[Any,Any,ResultType]]) -> None:
        self.function=function
        self.task=None
    async def __aenter__(self) -> asyncio.Task[ResultType]:
        '''create and return task'''
        assert self.task is None
        self.task=asyncio.create_task(self.function())
        return self.task
    async def __aexit__(self, t, e, b) -> None:
        '''cancel and await task'''
        assert self.task is not None
        task=self.task
        self.task=None
        if not task.done():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

class AsyncServiceQueue(contextlib.AbstractAsyncContextManager):
    loop:asyncio.AbstractEventLoop
    log_exception:Callable[[Exception],None]
    q:asyncio.Queue
    entered:bool

    def __init__(self,loop:asyncio.AbstractEventLoop,log_exception:Callable[[Exception],None]):
        self.loop=loop
        self.log_exception=log_exception
        self.q=asyncio.Queue()
        self.entered=False
        
    def enqueue(self, x:Coroutine[Any,Any,None]) -> None:
        '''enqueue {x} for later async execution
           - thread safe'''
        self.loop.call_soon_threadsafe(self.q.put_nowait,x)
        pass
    async def __aenter__(self) -> Self:
        self.entered=True
        return self
    async def __aexit__(self,t,e,b) -> None:
        self.entered=False
        # pop entries to avoid RuntimeWarning for coroutines never awaited noting
        # they were never started
        while not self.q.empty():
            x=self.q.get_nowait()
            assert isinstance(x,Coroutine)
            x.close()
        pass

    async def run(self) -> None:
        '''execute queue entries in order logging exceptions'''
        assert self.entered, f'AsyncServiceQueue context not entered'
        while True:
            x=await self.q.get()
            await self._execute(x)
        pass

    async def _execute(self,x:Coroutine[Any,Any,None]) -> None:
        try:
            await x
        except Exception:
            e=in_function_context(AsyncServiceQueue._execute,vars())
            assert isinstance(e,Exception)
            self.log_exception(e)
            pass
        pass
    pass

class AsyncMutex:
    def __init__(self):
        self.m = asyncio.Lock()
        pass
    pass

class AsyncLock(contextlib.AbstractAsyncContextManager):
    def __init__(self, m: AsyncMutex) -> None:
        self.m = m
        self.active=False
        pass

    async def __aenter__(self) -> Self:
        await self.m.m.acquire()
        self.active=True
        return self

    async def __aexit__(self,t,e,b) -> None:
        self.active=False
        self.m.m.release()
        pass
    pass

class AsyncCondition:
    def __init__(self, m:AsyncMutex):
        self.m = m
        self.c = asyncio.Condition(m.m)

    async def wait_for(self, l:AsyncLock, timeout:Duration):
        assert l.m == self.m and l.active
        async with asyncio.timeout(timeout.value()):
            await self.c.wait()
        pass

    def notify_all(self, l:AsyncLock):
        assert l.m == self.m and l.active
        self.c.notify_all()
    pass

async def delay_cancellation(t: Awaitable[T])->T:
    """await {t}, delaying any cancel of caller until {t} completes"""
    async def f():
        return await t
    async with AsyncTask(f) as task:
        really_cancel=task.cancel
        def delay_cancel(msg:Any=None):
            assert not task.cancelled()
            pass
        try:
            setattr(task,'cancel',delay_cancel)
            try:
                # Because we await task here, a cancel on us (i.e. caller)
                # will try to cancel task as well as us, which is why
                # we have overridden task's cancel method.
                # So task does not get cancelled, but we still do...
                return await task
            except asyncio.CancelledError:
                try:
                    await task
                except:
                    pass
                raise
        finally:
            setattr(task,'cancel',really_cancel)
            pass

def _make_base_classes_to_enter(
        cls: Type[ACMT]
) -> list[Callable[[ACMT],contextlib.AbstractAsyncContextManager[None]]]:
    result: list[Callable[[ACMT],contextlib.AbstractAsyncContextManager[None]]] = []
    for base_class in cls.__bases__:
        if base_class is not AsyncCM and issubclass(base_class, contextlib.AbstractAsyncContextManager):
            result.append(
                _make_async_class_cm(base_class))  # type: ignore  # above assert...
            # ... says that base_class is a subclass of AbstractAsyncContextManager i.e. an ACMT,
            # but mypy 1.4.1 says "Argument 1 to "_make_async_class_cm" has incompatible
            # type "type[AbstractAsyncContextManager[Any]]"; expected "type[ACMT]"
            pass
        assert not issubclass(cls, CM), cls
        pass
    return result

def _make_async_class_cm(
        cls:Type[ACMT]
) -> Callable[[ACMT],contextlib.AbstractAsyncContextManager[None]]:
    def result(x:contextlib.AbstractAsyncContextManager) -> contextlib.AbstractAsyncContextManager[None]:
        return _AsyncClassCm(cls, x)
    return result
            
@dataclass
class _AsyncClassCm(contextlib.AbstractAsyncContextManager[None]):
    '''target object {x} subclass {cls} context management methods
       - so for x: X where X(Y), __ClassCM(Y, x) will call Y.__enter__
         and Y.__exit__ directly avoiding any X.__enter__/__exit__
         overrides'''
    cls:Type[contextlib.AbstractAsyncContextManager[Any]]
    x:Any

    async def __aenter__(self) -> None:
        assert isinstance(self.x, self.cls)
        await self.cls.__aenter__(self.x)
    async def __aexit__(self, t, e, b):
        return await self.cls.__aexit__(self.x, t, e, b)
    pass

def _make_attrs_to_enter(
        cls: Type[ACMT]
) -> dict[
    str,
    Callable[
        [ACMT],contextlib.AbstractAsyncContextManager
    ]
]:
    result: dict[
        str,
        Callable[
            [ACMT],contextlib.AbstractAsyncContextManager
        ]
    ] = {}
    for n, t in cls.__annotations__.items():
        if _is_subclass(n, t, contextlib.AbstractAsyncContextManager):
            result[n]=_make_async_attr_class_cm(n)
        elif _is_subclass(n, t, contextlib.AbstractContextManager):
            result[n]=_make_async_sync_attr_class_cm(n)
            pass
        pass
    return result
    
def _make_async_attr_class_cm(
        attr_name: str
) -> Callable[[contextlib.AbstractAsyncContextManager[T]],contextlib.AbstractAsyncContextManager]:
    def result(x:contextlib.AbstractAsyncContextManager[T]) -> contextlib.AbstractAsyncContextManager:
        return _AsyncAttrClassCm(x, attr_name)
    return result
            
@dataclass
class _AsyncAttrClassCm(contextlib.AbstractAsyncContextManager):
    '''target class {self.x}.{self.attr_name} async context management methods'''
    x:contextlib.AbstractAsyncContextManager
    attr_name:str

    async def __aenter__(self):
        return await getattr(self.x, self.attr_name).__aenter__()
    async def __aexit__(self, t, e, b):
        return await getattr(self.x, self.attr_name).__aexit__(t, e, b)
        pass
    pass


def _make_async_sync_attr_class_cm(
        attr_name: str
) -> Callable[[contextlib.AbstractAsyncContextManager[T]],contextlib.AbstractAsyncContextManager]:
    def result(x:contextlib.AbstractAsyncContextManager[T]) -> contextlib.AbstractAsyncContextManager:
        return _AsyncSyncAttrClassCm(x, attr_name)
    return result

@dataclass
class _AsyncSyncAttrClassCm(contextlib.AbstractAsyncContextManager):
    '''target class {self.x}.{self.attr_name} non-async context management methods'''
    x:contextlib.AbstractAsyncContextManager
    attr_name:str

    async def __aenter__(self) -> None:
        return getattr(self.x, self.attr_name).__enter__()
    async def __aexit__(self, t, e, b) -> None:
        return getattr(self.x, self.attr_name).__exit__(t, e, b)
    pass



# Copyright (c) 2023 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# unix signal
#
from __future__ import annotations

from contextlib import AbstractContextManager
from asyncio import get_event_loop, Future, AbstractEventLoop
from xju import newtype
import signal

class SignumTag: pass
class Signum(newtype.Int[SignumTag]):
    pass

# some common signal types, get other values from signal module or OS
SIGTERM = Signum(signal.SIGTERM)
SIGPIPE = Signum(signal.SIGPIPE)

class AsyncSignalLatch(AbstractContextManager):
    loop: None|AbstractEventLoop
    result: None|Future[Signum]
    signum: Signum

    def __init__(self, signum: Signum):
        '''note must be created in main thread'''
        self.signum=signum
        pass

    def __enter__(self) -> Future[Signum]:
        '''latch next occurrance of signal {self.signum}'''
        self.loop = get_event_loop()
        self.result = self.loop.create_future()
        self.loop.add_signal_handler(self.signum.value(), self._signalled)
        return self.result

    def __exit__(self, t, e, b):
        assert self.result is not None
        assert self.loop is not None
        self.loop.remove_signal_handler(self.signum.value())
        

    def _signalled(self):
        assert self.result is not None
        self.result.set_result(self.signum)
        pass
    pass
